#!/usr/bin/env python3

import glob
import logging
import os
import subprocess as sp

logging.basicConfig(level=logging.INFO)

OUTPUT_PATH = "packages"


def replace_in_file(infile, replacements, outfile=None):
    """reads in a file, replaces the given data using python formatting and
    writes back the result to a file.

    Args:
        infile (str):
            File to be read
        replacements (dict):
            The replacements old => new in a dictionary format {old: new}
        outfile (str):
            Output file to which the data is written. If it is omitted, the
            input file will be overwritten instead

    """
    if outfile is None:
        outfile = infile

    with open(infile, "r") as fp:
        content = fp.read()

    for key, value in replacements.items():
        content = content.replace(key, value)

    with open(outfile, "w") as fp:
        fp.write(content)


def main(folder="droplets"):
    # remove old files
    for path in glob.glob(f"{OUTPUT_PATH}/*.rst"):
        logging.info("Remove file `%s`", path)
        os.remove(path)

    # run sphinx-apidoc
    sp.check_call(
        [
            "sphinx-apidoc",
            "--separate",
            "--maxdepth",
            "4",
            "--output-dir",
            OUTPUT_PATH,
            "--module-first",
            f"../../{folder}",  # path of the package
            f"../../{folder}/tests",  # ignored path
            f"../../{folder}/**/tests",  # ignored path
        ]
    )

    # replace unwanted information
    for path in glob.glob(f"{OUTPUT_PATH}/*.rst"):
        logging.info("Patch file `%s`", path)
        replace_in_file(path, {"Submodules\n----------\n\n": ""})


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""script that creates the resource with discretized surfaces of unit spheres """

import h5py
import numpy as np
import pygmsh

# generate all meshes
with h5py.File("spheres_3d.hdf5", "w") as f:
    num_list = set()

    # iterate over different characteristic lengths
    for length in [0.1, 0.2, 0.3, 0.5, 1.0, 2.0]:
        # initialize the geometry
        with pygmsh.occ.Geometry() as geom:
            # set the resolution of the geometry
            geom.characteristic_length_min = 0.75 * length
            geom.characteristic_length_max = length

            # add unit ball
            ball = geom.add_ball([0, 0, 0], 1.0)

            # generate mesh using gmsh
            mesh = geom.generate_mesh(dim=2, verbose=False)
            num = len(mesh.points)

            if num not in num_list:
                # save triangulated data to hdf file
                group = f.create_group(str(num))
                dset = group.create_dataset("points", data=mesh.points)
                dset.attrs["column_names"] = ["x", "y", "z"]

                # obtain angles
                φ = np.arctan2(mesh.points[:, 1], mesh.points[:, 0])
                r = np.hypot(mesh.points[:, 0], mesh.points[:, 1])
                θ = np.arctan2(r, mesh.points[:, 2])
                dset = group.create_dataset("angles", data=np.c_[φ, θ])
                dset.attrs["column_names"] = ["phi", "theta"]

                # cells
                cells = []
                for data in mesh.cells:
                    if data.type == "triangle":
                        cells.append(data.data)
                    else:
                        pass
                        # raise RuntimeError(f"Unsupported cell type `{data.type}`")
                cells = np.concatenate(cells)
                dset = group.create_dataset("cells", data=cells)
                dset.attrs["column_names"] = ["p1", "p2", "p3"]

                num_list.add(num)

    # store the number of generated spheres
    f.attrs["num_list"] = list(sorted(num_list))



"""
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

import numpy as np
import pytest
from numpy.lib.recfunctions import structured_to_unstructured

from pde import ScalarField
from pde.grids import (
    CartesianGrid,
    CylindricalSymGrid,
    PolarSymGrid,
    SphericalSymGrid,
    UnitGrid,
)

from droplets import image_analysis
from droplets.droplets import DiffuseDroplet, PerturbedDroplet2D, PerturbedDroplet3D
from droplets.emulsions import Emulsion


@pytest.mark.parametrize("size", [16, 17])
@pytest.mark.parametrize("periodic", [True, False])
def test_localization_sym_unit(size, periodic):
    """tests simple droplets localization in 2d"""
    pos = np.random.random(2) * size
    radius = np.random.uniform(2, 5)
    width = np.random.uniform(1, 2)
    d1 = DiffuseDroplet(pos, radius, interface_width=width)

    grid = UnitGrid((size, size), periodic=periodic)
    field = d1.get_phase_field(grid)

    emulsion = image_analysis.locate_droplets(field, refine=True)
    assert len(emulsion) == 1
    d2 = emulsion[0]

    np.testing.assert_almost_equal(d1.position, d2.position)
    assert d1.radius == pytest.approx(d2.radius, rel=1e-4)
    assert d1.interface_width == pytest.approx(d2.interface_width)

    emulsion = image_analysis.locate_droplets(field, minimal_radius=size)
    assert len(emulsion) == 0

    emulsion = image_analysis.locate_droplets(ScalarField(grid))
    assert len(emulsion) == 0


@pytest.mark.parametrize("periodic", [True, False])
def test_localization_sym_rect(periodic):
    """tests simple droplets localization in 2d with a rectangular grid"""
    size = 16

    pos = np.random.uniform(-4, 4, size=2)
    radius = np.random.uniform(2, 5)
    width = np.random.uniform(0.5, 1.5)
    d1 = DiffuseDroplet(pos, radius, interface_width=width)

    a = np.random.random(2) - size / 2
    b = np.random.random(2) + size / 2
    grid = CartesianGrid(np.c_[a, b], 3 * size, periodic=periodic)
    field = d1.get_phase_field(grid)

    emulsion = image_analysis.locate_droplets(field, refine=True)
    assert len(emulsion) == 1
    d2 = emulsion[0]

    np.testing.assert_almost_equal(d1.position, d2.position)
    assert d1.radius == pytest.approx(d2.radius, rel=1e-5)
    assert d1.interface_width == pytest.approx(d2.interface_width)

    emulsion = image_analysis.locate_droplets(ScalarField(grid))
    assert len(emulsion) == 0


@pytest.mark.parametrize("periodic", [True, False])
def test_localization_perturbed_2d(periodic):
    """tests localization of perturbed 2d droplets"""
    size = 16

    pos = np.random.uniform(-4, 4, size=2)
    radius = np.random.uniform(2, 5)
    width = np.random.uniform(0.5, 1.5)
    ampls = np.random.uniform(-0.01, 0.01, size=4)
    d1 = PerturbedDroplet2D(pos, radius, interface_width=width, amplitudes=ampls)

    a = np.random.random(2) - size / 2
    b = np.random.random(2) + size / 2
    grid = CartesianGrid(np.c_[a, b], 2 * size, periodic=periodic)
    field = d1.get_phase_field(grid)

    emulsion = image_analysis.locate_droplets(field, refine=True, modes=d1.modes)
    assert len(emulsion) == 1
    d2 = emulsion[0]

    msg = "size=%d, periodic=%s, %s != %s" % (size, periodic, d1, d2)
    np.testing.assert_almost_equal(d1.position, d2.position, decimal=1, err_msg=msg)
    assert d1.radius == pytest.approx(d2.radius, rel=1e-5)
    assert d1.interface_width == pytest.approx(d2.interface_width, rel=1e-3)
    np.testing.assert_allclose(d1.amplitudes[1:], d2.amplitudes[1:], rtol=0.5)


@pytest.mark.parametrize("periodic", [True, False])
def test_localization_perturbed_3d(periodic):
    """tests localization of perturbed 3d droplets"""
    size = 8

    pos = np.random.uniform(-2, 2, size=3)
    radius = np.random.uniform(2, 3)
    width = np.random.uniform(0.5, 1.5)
    ampls = np.random.uniform(-0.01, 0.01, size=3)
    d1 = PerturbedDroplet3D(pos, radius, interface_width=width, amplitudes=ampls)

    a = np.random.random(3) - size / 2
    b = np.random.random(3) + size / 2
    grid = CartesianGrid(np.c_[a, b], 2 * size, periodic=periodic)
    assert grid.dim == 3
    field = d1.get_phase_field(grid)

    emulsion = image_analysis.locate_droplets(field, refine=True, modes=d1.modes)
    assert len(emulsion) == 1
    d2 = emulsion[0]

    msg = "size=%d, periodic=%s, %s != %s" % (size, periodic, d1, d2)
    np.testing.assert_almost_equal(d1.position, d2.position, decimal=1, err_msg=msg)
    assert d1.radius == pytest.approx(d2.radius, rel=1e-4)
    assert d1.interface_width == pytest.approx(d2.interface_width, rel=1e-3)
    np.testing.assert_allclose(
        d1.amplitudes[3:], d2.amplitudes[3:], rtol=0.5, err_msg=msg
    )


def test_localization_polar():
    """tests simple droplets localization in polar grid"""
    radius = np.random.uniform(2, 3)
    width = np.random.uniform(0.5, 1.5)
    d1 = DiffuseDroplet((0, 0), radius, interface_width=width)

    grid_radius = 6 + 2 * np.random.random()
    grid = PolarSymGrid(grid_radius, 16)
    field = d1.get_phase_field(grid)

    emulsion = image_analysis.locate_droplets(field, refine=True)
    assert len(emulsion) == 1
    d2 = emulsion[0]

    np.testing.assert_almost_equal(d1.position, d2.position, decimal=5)
    assert d1.radius == pytest.approx(d2.radius, rel=1e-5)
    assert d1.interface_width == pytest.approx(d2.interface_width, rel=1e-5)

    emulsion = image_analysis.locate_droplets(ScalarField(grid))
    assert len(emulsion) == 0


def test_localization_spherical():
    """tests simple droplets localization in spherical grid"""
    radius = np.random.uniform(2, 3)
    width = np.random.uniform(0.5, 1.5)
    d1 = DiffuseDroplet((0, 0, 0), radius, interface_width=width)

    grid_radius = 6 + 2 * np.random.random()
    grid = SphericalSymGrid(grid_radius, 16)
    field = d1.get_phase_field(grid)

    emulsion = image_analysis.locate_droplets(field, refine=True)
    assert len(emulsion) == 1
    d2 = emulsion[0]

    np.testing.assert_almost_equal(d1.position, d2.position, decimal=5)
    assert d1.radius == pytest.approx(d2.radius, rel=1e-5)
    assert d1.interface_width == pytest.approx(d2.interface_width, rel=1e-5)

    emulsion = image_analysis.locate_droplets(ScalarField(grid))
    assert len(emulsion) == 0


@pytest.mark.parametrize("periodic", [True, False])
def test_localization_cylindrical(periodic):
    """tests simple droplets localization in cylindrical grid"""
    pos = (0, 0, np.random.uniform(-4, 4))
    radius = np.random.uniform(2, 3)
    width = np.random.uniform(0.5, 1.5)
    d1 = DiffuseDroplet(pos, radius, interface_width=width)

    grid_radius = 6 + 2 * np.random.random()
    bounds_z = np.random.uniform(1, 2, size=2) * np.array([-4, 4])
    grid = CylindricalSymGrid(grid_radius, bounds_z, (16, 32), periodic_z=periodic)
    field = d1.get_phase_field(grid)

    emulsion = image_analysis.locate_droplets(field, refine=True)
    assert len(emulsion) == 1
    d2 = emulsion[0]

    np.testing.assert_almost_equal(d1.position, d2.position, decimal=5)
    assert d1.radius == pytest.approx(d2.radius, rel=1e-5)
    assert d1.interface_width == pytest.approx(d2.interface_width)

    emulsion = image_analysis.locate_droplets(ScalarField(grid))
    assert len(emulsion) == 0


def test_localization_threshold():
    """tests different localization thresholds"""
    pos = np.random.random(2) * 16
    radius = np.random.uniform(2, 5)
    width = np.random.uniform(1, 2)
    d1 = DiffuseDroplet(pos, radius, interface_width=width)

    grid = UnitGrid((16, 16), periodic=False)
    field = d1.get_phase_field(grid)

    for threshold in [0.25, 0.75, "auto", "mean", "extrema", "otsu"]:
        emulsion = image_analysis.locate_droplets(
            field, threshold=threshold, refine=True
        )
        assert len(emulsion) == 1
        d2 = emulsion[0]

        np.testing.assert_almost_equal(d1.position, d2.position)
        assert d1.radius == pytest.approx(d2.radius, rel=1e-4)
        assert d1.interface_width == pytest.approx(d2.interface_width)


@pytest.mark.parametrize(
    "adjust_values, auto_values", [(False, False), (True, False), (True, True)]
)
def test_localization_vmin_vmax(adjust_values, auto_values):
    """tests localization of droplets with non-normalized densities"""
    # create perturbed droplet
    grid = CartesianGrid(bounds=[[-2, 2], [-2, 2]], shape=32, periodic=True)
    d1 = DiffuseDroplet([0, 0], 1, 0.2)
    field = d1.get_phase_field(grid, vmin=-0.1, vmax=0.1)

    if auto_values:
        refine_args = {"vmin": None, "vmax": None, "adjust_values": adjust_values}
    else:
        refine_args = {"vmin": -0.1, "vmax": 0.1, "adjust_values": adjust_values}

    # localize this droplet
    d2 = image_analysis.locate_droplets(
        field, threshold="auto", refine=True, refine_args=refine_args
    )[0]

    assert d1.position == pytest.approx(d2.position, rel=1e-4)
    assert d1.radius == pytest.approx(d2.radius, rel=1e-4)
    assert d1.interface_width == pytest.approx(d2.interface_width)


def test_get_structure_factor():
    """test the structure factor method"""
    grid = UnitGrid([512], periodic=True)
    k0 = np.random.uniform(1, 3)
    field = ScalarField.from_expression(grid, f"sin({k0} * x)")
    k, S = image_analysis.get_structure_factor(field)
    k_max = k[S.argmax()]
    assert k_max == pytest.approx(k0, rel=5e-1)


@pytest.mark.parametrize(
    "method",
    [
        "structure_factor_mean",
        "structure_factor_maximum",
        "droplet_detection",
    ],
)
def test_get_length_scale(method):
    """test determining the length scale"""
    grid = CartesianGrid([[0, 8 * np.pi]], 64, periodic=True)
    c = ScalarField.from_expression(grid, "sin(x)")
    s = image_analysis.get_length_scale(c, method=method)
    assert s == pytest.approx(2 * np.pi, rel=0.1)


def test_get_length_scale_edge():
    """test determining the length scale for edge cases"""
    grid = CartesianGrid(bounds=[[0, 1]], shape=32, periodic=True)
    for n in range(1, 4):
        c = ScalarField.from_expression(grid, f"0.2 + 0.2*sin(2*{n}*pi*x)")
        s = image_analysis.get_length_scale(c, method="structure_factor_maximum")
        assert s == pytest.approx(1 / n, rel=1e-4)


def test_emulsion_processing():
    """test identifying emulsions in phase fields"""
    grid = UnitGrid([32, 32], periodic=True)

    e1 = Emulsion(
        [
            DiffuseDroplet(position=[5, 6], radius=9, interface_width=1),
            DiffuseDroplet(position=[20, 19], radius=8, interface_width=1),
        ],
        grid=grid,
    )
    field = e1.get_phasefield()

    e2 = image_analysis.locate_droplets(field, refine=True)

    np.testing.assert_allclose(
        structured_to_unstructured(e1.data),
        structured_to_unstructured(e2.data),
        rtol=0.02,
    )


def test_structure_factor_random():
    """test the structure factor function for random input"""
    g1 = CartesianGrid([[0, 10]] * 2, 64, periodic=True)
    f1 = ScalarField.random_colored(g1, -2)

    # test invariance with re-meshing
    g2 = CartesianGrid([[0, 10]] * 2, [128, 64], periodic=True)
    f2 = f1.interpolate_to_grid(g2)

    ks = np.linspace(0.2, 3)
    k1, s1 = image_analysis.get_structure_factor(f1, wave_numbers=ks)
    k2, s2 = image_analysis.get_structure_factor(f2, wave_numbers=ks)

    np.testing.assert_equal(ks, k1)
    np.testing.assert_equal(ks, k2)
    np.testing.assert_allclose(s1, s2, atol=0.05)

    # test invariance with respect to scaling
    k2, s2 = image_analysis.get_structure_factor(100 * f1, wave_numbers=ks)
    np.testing.assert_equal(ks, k2)
    np.testing.assert_allclose(s1, s2, atol=0.05)



"""
Classes describing collections of droplets, i.e. emulsions, and their temporal dynamics.


.. autosummary::
   :nosignatures:

   Emulsion
   EmulsionTimeCourse


.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

from __future__ import annotations

import json
import logging
import math
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
    overload,
)

import numpy as np

from pde.fields import ScalarField
from pde.grids.base import GridBase
from pde.grids.cartesian import CartesianGrid
from pde.storage.base import StorageBase
from pde.tools.cuboid import Cuboid
from pde.tools.docstrings import fill_in_docstring
from pde.tools.output import display_progress
from pde.tools.plotting import PlotReference, plot_on_axes
from pde.trackers.base import InfoDict, IntervalData

from .droplets import SphericalDroplet, droplet_from_data

if TYPE_CHECKING:
    from .trackers import DropletTracker  # @UnusedImport


class Emulsion(list):
    """class representing a collection of droplets in a common system"""

    _show_projection_warning: bool = True
    """bool: Flag determining whether a warning is shown when high-dimensional
    emulsions are plotted"""

    grid: Optional[GridBase]

    def __init__(
        self,
        droplets: Optional[Iterable[SphericalDroplet]] = None,
        grid: Optional[GridBase] = None,
        copy: bool = True,
    ):
        """
        Args:
            droplets:
                A list or generator of instances of
                :class:`~droplets.droplets.SphericalDroplet`.
            grid (:class:`~pde.grids.base.GridBase`):
                The grid on which the droplets are defined. This information can
                helpful to measure distances between droplets correctly.
            copy (bool, optional):
                Whether to make a copy of the droplet or not
        """
        # obtain grid of the emulsion
        if isinstance(droplets, Emulsion):
            self.grid: Optional[GridBase] = droplets.grid
            if grid is not None and self.grid != grid:
                raise ValueError("Emulsion grid is unequal to given grid")
        else:
            self.grid = grid

        # determine space dimension
        if grid is not None:
            self.dim: Optional[int] = grid.dim  # dimension of the space
        else:
            self.dim = None

        # add all droplets
        super().__init__()
        if droplets is not None:
            self.extend(droplets, copy=copy)

    @classmethod
    def from_random(
        cls,
        num: int,
        grid_or_bounds: Union[GridBase, Sequence[Tuple[float, float]]],
        radius: Union[float, Tuple[float, float]],
        *,
        remove_overlapping: bool = True,
        droplet_class: Type[SphericalDroplet] = SphericalDroplet,
        rng: Optional[np.random.Generator] = None,
    ) -> Emulsion:
        """
        Create an emulsion with random droplets

        Args:
            num (int):
                The (maximal) number of droplets to generate
            grid_or_bounds (:class:`~pde.grids.base.GridBase` or list of float tuples):
                Determines the space in which droplets are placed. This is either a
                :class:`~pde.grids.base.GridBase` describing the geometry or a sequence
                of tuples with lower and upper bounds for each axes, so the length of
                the sequence determines the space dimension.
            radius (float or tuple of float):
                Radius of the droplets that are created. If two numbers are given, they
                specify the bounds of a uniform distribution from which the radius of
                each individual droplet is chosen.
            remove_overlapping (bool):
                Flag determining whether overlapping droplets are removed. If enabled,
                the resulting element might contain less thatn `num` droplets.
            droplet_class:
                The class that is used to describe droplets.
            rng (:class:`~numpy.random.Generator`):
                Random number generator (default: :func:`~numpy.random.default_rng()`)
        """
        if rng is None:
            rng = np.random.default_rng()

        # determine how to get random droplet positions
        if isinstance(grid_or_bounds, GridBase):

            def get_position():
                return grid_or_bounds.get_random_point(rng=rng)

        else:
            bnds = np.atleast_2d(grid_or_bounds)
            assert bnds.ndim == 2 and bnds.shape[0] > 0 and bnds.shape[1] == 2

            def get_position():
                return rng.uniform(bnds[:, 0], bnds[:, 1])

        # extract information about radius
        try:
            r0, r1 = radius  # type: ignore
        except TypeError:
            r0 = r1 = float(radius)  # type: ignore

        # create the emulsion from a list of droplets
        drops = [droplet_class(get_position(), rng.uniform(r0, r1)) for _ in range(num)]
        emulsion = cls(drops)

        if remove_overlapping:
            emulsion.remove_overlapping()

        return emulsion

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return super().__eq__(other) and self.grid == other.grid

    def __ne__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return super().__ne__(other) or self.grid != other.grid

    def __repr__(self):
        return f"{self.__class__.__name__}({super().__repr__()}, grid={self.grid})"

    def __add__(self, rhs):
        # return sum of Emulsions as an emulsion
        return Emulsion(list.__add__(self, rhs))

    @overload  # type: ignore
    def __getitem__(self, key: int) -> SphericalDroplet:
        ...

    @overload
    def __getitem__(self, key: slice) -> Emulsion:
        ...

    def __getitem__(self, key):
        # return result from extended slicing as Emulsion
        result = list.__getitem__(self, key)
        if isinstance(result, SphericalDroplet):
            return result
        else:
            return Emulsion(result, grid=self.grid)

    def copy(self, min_radius: float = -1) -> Emulsion:
        """return a copy of this emulsion

        Args:
            min_radius (float):
                The minimal radius of the droplets that are retained. Droplets with
                exactly min_radius are removed, so `min_radius == 0` can be used to
                filter vanished droplets.
        """
        droplets: List[SphericalDroplet] = [
            droplet.copy() for droplet in self if droplet.radius > min_radius
        ]
        return self.__class__(
            droplets,
            grid=self.grid,
            copy=False,
        )

    def extend(
        self,
        droplets: Iterable[SphericalDroplet],
        copy: bool = True,
    ) -> None:
        """add many droplets to the emulsion

        Args:
            droplets (list): List of droplets to add to the emulsion
            copy (bool, optional): Whether to make a copy of the droplets or not
        """
        for droplet in droplets:
            self.append(droplet, copy=copy)

    def append(self, droplet: SphericalDroplet, copy: bool = True) -> None:
        """add a droplet to the emulsion

        Args:
            droplet (:class:`droplets.dropelts.SphericalDroplet`):
                Droplet to add to the emulsion
            copy (bool, optional):
                Whether to make a copy of the droplet or not
        """
        if self.dim is None:
            self.dim = droplet.dim
        elif self.dim != droplet.dim:
            raise ValueError(
                f"Cannot append droplet in dimension {droplet.dim} to emulsion in "
                f"dimension {self.dim}"
            )
        if copy:
            droplet = droplet.copy()
        super().append(droplet)

    @property
    def data(self) -> Optional[np.ndarray]:
        """:class:`~numpy.ndarray`: an array containing the data of the full emulsion

        Warning:
            This requires all droplets to be of the same class. The returned array is
            a copy of all the data and writing to it will thus not change the underlying
            data.
        """
        if len(self) == 0:
            return None
        else:
            # emulsion contains at least one droplet
            classes = set(d.__class__ for d in self)
            if len(classes) > 1:
                raise TypeError(
                    "Emulsion data cannot be stored contiguously if it contains a "
                    f"multiple of droplet classes: "
                    + ", ".join(c.__name__ for c in classes)
                )
            return np.array([d.data for d in self])

    def get_linked_data(self) -> np.ndarray:
        """link the data of all droplets in a single array

        Returns:
            :class:`~numpy.ndarray`: The array containing all droplet data. If entries in
                this array are modified, it will be reflected in the droplets.
        """
        if len(self) == 0:
            data = np.empty(0)
        else:
            assert self.data is not None
            data = self.data  # create an array with all the droplet data
            # link back to droplets
            for i, d in enumerate(self):
                d.data = data[i]
        return data

    @classmethod
    def _from_hdf_dataset(cls, dataset, grid: Optional[GridBase] = None) -> Emulsion:
        """construct an emulsion by reading data from an hdf5 dataset

        Args:
            dataset:
                An HDF5 dataset from which the data of the emulsion is read
            grid (:class:`pde.grids.base.GridBase`):
                The grid on which the droplets are defined. This information is required
                to measure distances between droplets.
        """
        # there are values, so the emulsion is not empty
        droplet_class = dataset.attrs["droplet_class"]
        if droplet_class == "None":
            droplets: List[SphericalDroplet] = []
        else:
            droplets = [
                droplet_from_data(droplet_class, data)  # type: ignore
                for data in dataset
            ]
        return cls(droplets, grid=grid, copy=False)

    @classmethod
    def from_file(cls, path: str) -> Emulsion:
        """create emulsion by reading file

        Args:
            path (str):
                The path from which the data is read. This function assumes that the
                data was written as an HDF5 file using :meth:`to_file`.
        """
        import h5py

        with h5py.File(path, "r") as fp:
            if len(fp) != 1:
                raise RuntimeError(f"Multiple emulsions found in file `{path}`")

            # read grid
            if "grid" in fp.attrs:
                grid: Optional[GridBase] = GridBase.from_state(fp.attrs["grid"])
            else:
                grid = None

            # read the actual droplet data
            dataset = fp[list(fp.keys())[0]]  # retrieve the only dataset
            obj = cls._from_hdf_dataset(dataset, grid)

        return obj

    def _write_hdf_dataset(self, hdf_path, key: str = "emulsion"):
        """write data to a given hdf5 path `hdf_path`"""
        if self:
            # emulsion contains at least one droplet
            dataset = hdf_path.create_dataset(key, data=self.data)
            # self.data ensures that there is only one droplet class
            dataset.attrs["droplet_class"] = self[0].__class__.__name__

        else:
            # create empty dataset to indicate empty emulsion
            dataset = hdf_path.create_dataset(key, shape=tuple())
            dataset.attrs["droplet_class"] = "None"

        return dataset

    def to_file(self, path: str) -> None:
        """store data in hdf5 file

        The data can be read using the classmethod :meth:`Emulsion.from_file`.

        Args:
            path (str):
                The path to which the data is written as an HDF5 file.
        """
        import h5py

        with h5py.File(path, "w") as fp:
            # write the grid data
            if self.grid is not None:
                fp.attrs["grid"] = self.grid.state_serialized
            # write actual droplet data
            self._write_hdf_dataset(fp)

    @property
    def interface_width(self) -> Optional[float]:
        """float: the average interface width across all droplets

        This averages the interface widths of the individual droplets weighted by their
        surface area, i.e., the amount of interface.
        """

        width, area = 0.0, 0.0
        for droplet in self:
            try:
                interface_width = droplet.interface_width
            except AttributeError:
                pass
            else:
                if interface_width is not None:
                    a = droplet.surface_area
                    width += interface_width * a
                    area += a

        if area == 0:
            return None
        else:
            return width / area

    @property
    def bbox(self) -> Cuboid:
        """:class:`Cuboid`: bounding box of the emulsion"""
        if len(self) == 0:
            raise RuntimeError("Bounding box of empty emulsion is undefined")
        return sum((droplet.bbox for droplet in self[1:]), self[0].bbox)

    def get_phasefield(
        self, grid: Optional[GridBase] = None, label: Optional[str] = None
    ) -> ScalarField:
        """create a phase field representing a list of droplets

        Args:
            grid (:class:`pde.grids.base.GridBase`):
                The grid on which the phase field is created. If omitted, the grid
                associated with the emulsion is used.
            label (str):
                Optional label for the returned scalar field

        Returns:
            :class:`~pde.fields.scalar.ScalarField`: the actual phase field
        """
        if grid is None:
            grid = self.grid
        if grid is None:
            raise RuntimeError("Grid needs to be specified")

        if len(self) == 0:
            return ScalarField(grid)

        else:
            result: ScalarField = self[0].get_phase_field(grid, label=label)
            for d in self[1:]:
                result += d.get_phase_field(grid)
            np.clip(result.data, 0, 1, out=result.data)
            return result

    def remove_small(self, min_radius: float = -np.inf) -> None:
        """remove droplets that are very small

        The emulsions is modified in-place.

        Args:
            min_radius (float):
                The minimal radius of the droplets that are retained. Droplets with
                exactly min_radius are removed, so `min_radius == 0` can be used to
                filter vanished droplets. The default value does not remove any droplets
        """
        for i in reversed(range(len(self))):
            if self[i].radius <= min_radius:
                self.pop(i)

    def get_pairwise_distances(self, subtract_radius: bool = False) -> np.ndarray:
        """return the pairwise distance between droplets

        Args:
            subtract_radius (bool):
                determines whether the distance is measured from interface to interface
                (for round droplets) or center to center.

        Returns:
            :class:`~numpy.ndarray`: a matrix with the distances between all droplets
        """
        if self.grid is None:

            def get_distance(p1, p2):
                """helper function calculating the distance between points"""
                return np.linalg.norm(p1 - p2)

        else:
            get_distance = self.grid.distance_real

        # calculate pairwise distance and return it in requested form
        num = len(self)
        dists = np.zeros((num, num))
        # iterate over all droplet pairs
        for i in range(num):
            for j in range(i + 1, num):
                d1, d2 = self[i], self[j]
                dist = get_distance(d1.position, d2.position)
                if subtract_radius:
                    dist -= d1.radius + d2.radius
                dists[i, j] = dists[j, i] = dist

        return dists

    def get_neighbor_distances(self, subtract_radius: bool = False) -> np.ndarray:
        """calculates the distance of each droplet to its nearest neighbor

        Warning:
            Nearest neighbors are defined by comparing the distances between the centers
            of the droplets, not their surfaces.

        Args:
            subtract_radius (bool):
                Determines whether to subtract the radius from the distance, i.e.,
                whether to return the distance between the surfaces instead of the
                positions

        Returns:
            :class:`~numpy.ndarray`: a vector with a distance for each droplet
        """
        # handle simple cases
        if len(self) == 0:
            return np.zeros((0,))
        elif len(self) == 1:
            return np.full(1, np.nan)

        try:
            from scipy.spatial import cKDTree as KDTree
        except ImportError:
            from scipy.spatial import KDTree

        # build tree to query the nearest neighbors
        assert self.data is not None
        positions = self.data["position"]
        tree = KDTree(positions)
        dist, index = tree.query(positions, 2)

        if subtract_radius:
            return dist[:, 1] - self.data["radius"][index].sum(axis=1)  # type: ignore
        else:
            return dist[:, 1]  # type: ignore

    def remove_overlapping(self, min_distance: float = 0) -> None:
        """remove all droplets that are overlapping

        If a pair of overlapping droplets was found, the smaller one of these is removed
        from the current emulsion. This method modifies the emulsion in place and thus
        does not return anything.

        Args:
            min_distance (float):
                The minimal distance droplets need to be apart. The default value of `0`
                corresponds to just remove overlapping droplets. Larger values ensure
                that droplets keep a distance, while negative values allow for some
                overlap.
        """
        # filter duplicates until there are none left
        dists = self.get_pairwise_distances(subtract_radius=True)
        np.fill_diagonal(dists, np.inf)

        while len(dists) > 1:
            # find minimal distance
            x, y = np.unravel_index(np.argmin(dists), dists.shape)
            if dists[x, y] < min_distance:
                # droplets overlap -> remove the smaller one
                if self[x].radius > self[y].radius:  # type: ignore
                    self.pop(y)
                    dists = np.delete(np.delete(dists, y, 0), y, 1)
                else:
                    self.pop(x)
                    dists = np.delete(np.delete(dists, x, 0), x, 1)
            else:
                break

    @property
    def total_droplet_volume(self) -> float:
        """float: the total volume of all droplets"""
        return sum(droplet.volume for droplet in self)  # type: ignore

    def get_size_statistics(self, incl_vanished: bool = True) -> Dict[str, float]:
        """determine size statistics of the current emulsion

        Args:
            incl_vanished (bool):
                Whether to include droplets with vanished radii

        Returns:
            dict: a dictionary with various size statistics
        """
        if len(self) == 0:
            return {
                "count": 0,
                "radius_mean": math.nan,
                "radius_std": math.nan,
                "volume_mean": math.nan,
                "volume_std": math.nan,
            }

        if incl_vanished:
            radii = [droplet.radius for droplet in self]
            volumes = [droplet.volume for droplet in self]
        else:
            radii = [droplet.radius for droplet in self if droplet.radius > 0]
            volumes = [droplet.volume for droplet in self if droplet.radius > 0]

        return {
            "count": len(radii),
            "radius_mean": np.mean(radii),
            "radius_std": np.std(radii),
            "volume_mean": np.mean(volumes),
            "volume_std": np.std(volumes),
        }

    @plot_on_axes()
    def plot(
        self,
        ax,
        field: Optional[ScalarField] = None,
        image_args: Optional[Dict[str, Any]] = None,
        repeat_periodically: bool = True,
        color_value: Optional[Callable] = None,
        cmap=None,
        norm=None,
        colorbar: Union[bool, str] = True,
        **kwargs,
    ) -> PlotReference:
        """plot the current emulsion together with a corresponding field

        If the emulsion is defined in a 3d geometry, only a projection on the first two
        axes is shown.

        Args:
            {PLOT_ARGS}
            field (:class:`pde.fields.scalar.ScalarField`):
                provides the phase field that is shown as a background
            image_args (dict):
                additional arguments determining how the phase field in the
                background is plotted. Acceptable arguments are described in
                :func:`~pde.fields.base.FieldBase.plot`.
            repeat_periodically (bool):
                flag determining whether droplets are shown on both sides of
                periodic boundary conditions. This option can slow down plotting
            color_value (callable):
                Function used to determine the color of a droplet. The function is
                called with individual droplet objects and must return a single scalar
                value, which is then mapped to a color using the colormap given by
                `cmap` and a suitable normalization given by `norm`.
            cmap (str or :class:`~matplotlib.colors.Colormap`):
                The colormap used to map normalized data values to RGBA colors.
            norm (:class:`~matplotlib.colors.Normalize`):
                The normalizing object which scales data, typically into the interval
                [0, 1]. If None, norm defaults to a `colors.Normalize` object which
                maps the range of values obtained from `color_value` to [0, 1].
            colorbar (bool or str):
                Determines whether a colorbar is shown when `color_value` is supplied.
                If a string is given, it is used as a label for the colorbar.
            **kwargs:
                Additional keyword arguments are passed to the function creating the
                patch that represents the droplet. For instance, to only draw the
                outlines of the droplets, you may need to supply `fill=False`.

        Returns:
            :class:`~pde.tools.plotting.PlotReference`: Information about the plot
        """
        if len(self) == 0:
            # empty emulsions can be plotted in all dimensions :)
            return PlotReference(ax, [], {})
        if self.dim is None or self.dim <= 1:
            raise NotImplementedError(
                f"Plotting emulsions in {self.dim} dimensions is not implemented."
            )
        elif self.dim > 2:
            if Emulsion._show_projection_warning:
                logger = logging.getLogger(self.__class__.__name__)
                logger.warning("A projection on the first two axes is shown.")
                Emulsion._show_projection_warning = False

        grid_compatible = (
            self.grid is None or field is None or self.grid.compatible_with(field.grid)
        )
        if not grid_compatible:
            raise ValueError("Emulsion grid is not compatible with field grid")
        grid = self.grid

        # plot background and determine bounds for the droplets
        if field is not None:
            # plot the phase field and use its bounds
            if image_args is None:
                image_args = {}
            field.plot(kind="image", ax=ax, **image_args)
            ax.autoscale(False)  # fix image bounds to phase field
            if grid is None:
                grid = field.grid

        else:
            if isinstance(grid, CartesianGrid):
                # determine the bounds from the (2d) grid
                bounds = grid.axes_bounds
            else:
                # determine the bounds from the emulsion data itself
                bounds = self.bbox.bounds
            ax.set_xlim(*bounds[0])
            ax.set_ylim(*bounds[1])
            ax.set_aspect("equal")

        # determine non-vanishing droplets
        drops_finite = [droplet for droplet in self if droplet.radius > 0]

        # determine the color of all droplets
        if color_value is not None:
            import matplotlib.pyplot as plt

            # determine the scalar values associated with all droplets
            values = np.array([color_value(droplet) for droplet in drops_finite])

            # and map them to colors
            mapper = plt.cm.ScalarMappable(norm=norm, cmap=cmap)
            colors = mapper.to_rgba(values)

            if kwargs.pop("color", None) is not None:
                logger = logging.getLogger(self.__class__.__name__)
                logger.warning("`color` is overwritten by `color_value`.")

        else:
            colors = [kwargs.pop("color", None)] * len(drops_finite)

        # get patches representing all droplets
        if grid is None or not repeat_periodically:
            # plot only the droplets themselves
            patches = [
                droplet._get_mpl_patch(dim=2, color=color, **kwargs)
                for droplet, color in zip(drops_finite, colors)
            ]
        else:
            # plot droplets also in their mirror positions
            patches = []
            for droplet, color in zip(drops_finite, colors):
                for p in grid.iter_mirror_points(
                    droplet.position, with_self=True, only_periodic=True
                ):
                    # create copy with changed position
                    d = droplet.copy(position=p)
                    patches.append(d._get_mpl_patch(dim=2, color=color, **kwargs))

        # add all patches as a collection
        import matplotlib as mpl

        # the zorder needs to be set on the collection level to have any effect
        col_args = {}
        if "zorder" in kwargs:
            col_args["zorder"] = kwargs["zorder"]
        coll = mpl.collections.PatchCollection(patches, match_original=True, **col_args)
        ax.add_collection(coll)

        # add colorbar if requested
        if color_value is not None and colorbar:
            from pde.tools.plotting import add_scaled_colorbar

            label = colorbar if isinstance(colorbar, str) else ""
            add_scaled_colorbar(mapper, ax=ax, label=label)

        parameters = {
            "repeat_periodicially": repeat_periodically,
        }
        return PlotReference(ax, coll, parameters)


class EmulsionTimeCourse:
    """represents emulsions as a function of time"""

    def __init__(
        self,
        emulsions: Optional[Iterable[Emulsion]] = None,
        times: Union[np.ndarray, Sequence[float], None] = None,
    ) -> None:
        """
        Args:
            emulsions (list): List of emulsions that describe this time course
            times (list): Times associated with the emulsions
        """
        if isinstance(emulsions, EmulsionTimeCourse):
            # extract data from given object; ignore `times`
            times = emulsions.times
            self.grid: Optional[GridBase] = emulsions.grid
            emulsions = emulsions.emulsions
        else:
            self.grid = None

        self.emulsions: List[Emulsion] = []
        self.times: List[float] = []

        # add all emulsions
        if emulsions is not None:
            for e in emulsions:
                self.append(Emulsion(e))

        # add all times
        if times is None:
            self.times = list(range(len(self.emulsions)))
        else:
            self.times = list(times)

        if len(self.times) != len(self.emulsions):
            raise ValueError("Lists of emulsions and times must have same length")

    def append(
        self, emulsion: Emulsion, time: Optional[float] = None, copy: bool = True
    ) -> None:
        """add an emulsion to the list

        Args:
            emulsions (Emulsion):
                An :class:`Emulsion` instance that is added to the time course
            time (float):
                The time point associated with this emulsion
            copy (bool):
                Whether to copy the emulsion
        """
        emulsion = Emulsion(emulsion)  # make sure this is an emulsion

        # check grid consistency
        if self.grid is None:
            self.grid = emulsion.grid
        elif emulsion.grid and not self.grid.compatible_with(emulsion.grid):
            raise ValueError(
                "Grid of the EmulsionTimeCourse is not compatible with the grid of the "
                f"emulsion to be added. ({self.grid} != {emulsion.grid})"
            )
        # add the emulsion
        if copy:
            emulsion = emulsion.copy()
        self.emulsions.append(emulsion)

        if time is None:
            time = 0 if len(self.times) == 0 else self.times[-1] + 1
        self.times.append(time)

    def clear(self) -> None:
        """removes all data stored in this instance"""
        self.emulsions = []
        self.times = []

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(emulsions=Array({len(self)}), "
            f"times=Array({len(self)}), grid={self.grid})"
        )

    def __len__(self):
        return len(self.times)

    def __getitem__(self, key: Union[int, slice]):
        """return the information for the given index"""
        result = self.emulsions.__getitem__(key)
        if isinstance(key, slice):
            return self.__class__(emulsions=result, times=self.times[key])
        else:
            return result

    def __iter__(self) -> Iterator[Emulsion]:
        """iterate over the emulsions"""
        return iter(self.emulsions)

    def items(self) -> Iterator[Tuple[float, Emulsion]]:
        """iterate over all times and emulsions, returning them in pairs"""
        return zip(self.times, self.emulsions)

    def __eq__(self, other):
        """determine whether two EmulsionTimeCourse instance are equal"""
        grids_equal = self.grid is None or other.grid is None or self.grid == other.grid
        return (
            grids_equal
            and self.times == other.times
            and self.emulsions == other.emulsions
        )

    @classmethod
    def from_storage(
        cls,
        storage: StorageBase,
        refine: bool = False,
        progress: Optional[bool] = None,
        **kwargs,
    ) -> "EmulsionTimeCourse":
        r"""create an emulsion time course from a stored phase field

        Args:
            storage (:class:`~pde.storage.base.StorageBase`):
                The phase fields for many time instances
            refine (bool):
                Flag determining whether the droplet properties should be refined
                using fitting. This is a potentially slow procedure.
            progress (bool):
                Whether to show the progress of the process. If `None`, the progress is
                only shown when `refine` is `True`.
            \**kwargs:
                All other parameters are forwarded to the
                :meth:`~droplets.image_analysis.locate_droplets`.

        Returns:
            EmulsionTimeCourse: an instance describing the emulsion time course
        """
        from .image_analysis import locate_droplets

        if progress is None:
            progress = refine  # show progress only when refining by default

        # obtain the emulsion data for all frames
        emulsions = (
            locate_droplets(frame, refine=refine, **kwargs)
            for frame in display_progress(storage, enabled=progress)
        )

        return cls(emulsions, times=storage.times)

    @classmethod
    def from_file(cls, path: str, progress: bool = True) -> "EmulsionTimeCourse":
        """create emulsion time course by reading file

        Args:
            path (str):
                The path from which the data is read. This function assumes that the
                data was written as an HDF5 file using :meth:`to_file`.
            progress (bool):
                Whether to show the progress of the process in a progress bar

        Returns:
            EmulsionTimeCourse: an instance describing the emulsion time course
        """
        import h5py

        obj = cls()
        with h5py.File(path, "r") as fp:
            # read grid
            if "grid" in fp.attrs:
                grid: Optional[GridBase] = GridBase.from_state(fp.attrs["grid"])
            else:
                grid = None

            # load the actual emulsion data and iterate in the right order
            for key in display_progress(
                sorted(fp.keys()), total=len(fp), enabled=progress
            ):
                dataset = fp[key]
                obj.append(
                    Emulsion._from_hdf_dataset(dataset, grid),
                    time=dataset.attrs["time"],
                )
        return obj

    def to_file(self, path: str, info: Optional[InfoDict] = None) -> None:
        """store data in hdf5 file

        The data can be read using the classmethod :meth:`EmulsionTimeCourse.from_file`.

        Args:
            path (str):
                The path to which the data is written as an HDF5 file.
            info (dict):
                Additional data stored alongside the droplet track list
        """
        import h5py

        with h5py.File(path, "w") as fp:
            # write the actual emulsion data
            for i, (time, emulsion) in enumerate(self.items()):
                dataset = emulsion._write_hdf_dataset(fp, f"time_{i:06d}")
                dataset.attrs["time"] = time

            # write additional information
            if info:
                for k, v in info.items():
                    fp.attrs[k] = json.dumps(v)

            # write the grid data -> this might overwrite grid data that is
            # present in the info dictionary, but in normal cases these grids
            # should be identical, so we don't handle this case explicitly
            if self.grid is not None:
                fp.attrs["grid"] = self.grid.state_serialized

    def get_emulsion(self, time: float) -> Emulsion:
        """returns the emulsion clostest to a specific time point

        Args:
            time (float): The time point

        Returns:
            :class:`Emuslion`
        """
        idx = np.argmin(np.abs(np.asarray(self.times) - time))
        return self.emulsions[idx]

    @fill_in_docstring
    def tracker(
        self, interval: IntervalData = 1, filename: Optional[str] = None
    ) -> "DropletTracker":
        """return a tracker that analyzes emulsions during simulations

        Args:
            interval:
                {ARG_TRACKER_INTERVAL}
            filename (str): determines where the EmulsionTimeCourse data is
                stored
        """
        from .trackers import DropletTracker  # @Reimport

        return DropletTracker(
            emulsion_timecourse=self, filename=filename, interval=interval
        )


__all__ = ["Emulsion", "EmulsionTimeCourse"]



#!/usr/bin/env python3

import argparse
import os
import subprocess as sp
import sys
from pathlib import Path
from typing import Sequence

PACKAGE = "droplets"  # name of the package that needs to be tested
PACKAGE_PATH = Path(__file__).resolve().parents[1]  # base path of the package


def _most_severe_exit_code(retcodes: Sequence[int]) -> int:
    """returns the most severe exit code of a given list

    Args:
        retcodes (list): A list of return codes

    Returns:
        int: the exit code that is most severe
    """
    if all(retcode == 0 for retcode in retcodes):
        return 0
    else:
        return max(retcodes, key=lambda retcode: abs(retcode))


def test_codestyle(*, verbose: bool = True) -> int:
    """run the codestyle tests

    Args:
        verbose (bool): Whether to do extra output

    Returns:
        int: The most severe exit code
    """
    retcodes = []

    for folder in [PACKAGE, "examples"]:
        if verbose:
            print(f"Checking codestyle in folder {folder}...")
        path = PACKAGE_PATH / folder

        # format imports
        result = sp.run(["isort", "--diff", path])
        retcodes.append(result.returncode)
        # format rest
        result = sp.run(["black", "--check", path])
        retcodes.append(result.returncode)

    return _most_severe_exit_code(retcodes)


def test_types(*, report: bool = False, verbose: bool = True) -> int:
    """run mypy to check the types of the python code

    Args:
        report (bool): Whether to write a report
        verbose (bool): Whether to do extra output

    Returns:
        int: The return code indicating success or failure
    """
    if verbose:
        print(f"Checking types in the {PACKAGE} package...")

    args = [
        sys.executable,
        "-m",
        "mypy",
        "--config-file",
        PACKAGE_PATH / "pyproject.toml",
    ]

    if report:
        folder = PACKAGE_PATH / "tests" / "mypy-report"
        if verbose:
            print(f"Writing report to `{folder}`")
        args.extend(
            ["--no-incremental", "--linecount-report", folder, "--html-report", folder]
        )
    else:
        # do not create a report
        args.append("--pretty")

    args.extend(["--package", PACKAGE])

    return sp.run(args, cwd=PACKAGE_PATH).returncode


def run_unit_tests(
    runslow: bool = False,
    parallel: bool = False,
    coverage: bool = False,
    no_numba: bool = False,
    pattern: str = None,
) -> int:
    """run the unit tests

    Args:
        runslow (bool): Whether to run the slow tests
        parallel (bool): Whether to use multiple processors
        coverage (bool): Whether to determine the test coverage
        no_numba (bool): Whether to disable numba jit compilation
        pattern (str): A pattern that determines which tests are ran

    Returns:
        int: The return code indicating success or failure
    """
    # modify current environment
    env = os.environ.copy()
    env["MPLBACKEND"] = "agg"
    if no_numba:
        env["NUMBA_DISABLE_JIT"] = "1"
    else:
        env["NUMBA_WARNINGS"] = "1"
        env["NUMBA_BOUNDSCHECK"] = "1"

    # build the arguments string
    args = [
        sys.executable,
        "-m",
        "pytest",  # run pytest module
        "-c",
        "pyproject.toml",  # locate the configuration file
        "-rs",  # show summary of skipped tests
        "-rw",  # show summary of warnings raised during tests
    ]

    # allow running slow tests?
    if runslow:
        args.append("--runslow")

    # run tests using multiple cores?
    if parallel:
        from multiprocessing import cpu_count

        args.extend(["-n", str(cpu_count() // 2), "--durations=10"])

    # run only a subset of the tests?
    if pattern is not None:
        args.extend(["-k", str(pattern)])

    # add coverage attributes?
    if coverage:
        args.extend(
            [
                "--cov-config=pyproject.toml",  # locate the configuration file
                "--cov-report",
                "html:tests/coverage",  # create a report in html format
                f"--cov={PACKAGE}",  # specify in which package the coverage is measured
            ]
        )

    # specify the package to run
    args.append("tests")

    # actually run the test
    return sp.run(args, env=env, cwd=PACKAGE_PATH).returncode


def main() -> int:
    """the main program controlling the tests

    Returns:
        int: The return code indicating success or failure
    """
    # parse the command line arguments
    parser = argparse.ArgumentParser(
        description=f"Run tests of the `{PACKAGE}` package.",
        epilog="All test categories are run if no specific categories are selected.",
    )

    # add the basic tests that need to be run
    group = parser.add_argument_group("Test categories")
    group.add_argument(
        "-s", "--style", action="store_true", default=False, help="Test code style"
    )
    group.add_argument(
        "-t", "--types", action="store_true", default=False, help="Test object types"
    )
    group.add_argument(
        "-u", "--unit", action="store_true", default=False, help="Run unit tests"
    )

    # set additional arguments
    group = parser.add_argument_group("Additional arguments")
    group.add_argument(
        "-q",
        "--quite",
        action="store_true",
        default=False,
        help="Suppress output from the script",
    )
    group.add_argument(
        "--runslow",
        action="store_true",
        default=False,
        help="Also run slow unit tests",
    )
    group.add_argument(
        "--coverage",
        action="store_true",
        default=False,
        help="Record test coverage of unit tests",
    )
    group.add_argument(
        "--parallel",
        action="store_true",
        default=False,
        help="Use multiprocessing",
    )
    group.add_argument(
        "--no_numba",
        action="store_true",
        default=False,
        help="Do not use just-in-time compilation of numba",
    )
    group.add_argument(
        "--pattern",
        metavar="PATTERN",
        type=str,
        help="Only run tests with this pattern",
    )
    group.add_argument(
        "--report",
        action="store_true",
        default=False,
        help="Write a report of the results",
    )

    # parse the command line arguments
    args = parser.parse_args()
    run_all = not (args.style or args.types or args.unit)

    # run the requested tests
    retcodes = []
    if run_all or args.style:
        retcode = test_codestyle(verbose=not args.quite)
        retcodes.append(retcode)

    if run_all or args.types:
        retcode = test_types(report=args.report, verbose=not args.quite)
        retcodes.append(retcode)

    if run_all or args.unit:
        retcode = run_unit_tests(
            runslow=args.runslow,
            coverage=args.coverage,
            parallel=args.parallel,
            no_numba=args.no_numba,
            pattern=args.pattern,
        )
        retcodes.append(retcode)

    # return the most severe code
    return _most_severe_exit_code(retcodes)


if __name__ == "__main__":
    sys.exit(main())



"""
Functions for analyzing phase field images of emulsions.

.. autosummary::
   :nosignatures:

   locate_droplets
   refine_droplet
   get_structure_factor
   get_length_scale

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

import logging
import math
import warnings
from functools import reduce
from typing import Any, Dict, Optional, Sequence, Tuple, Union

import numpy as np
from numpy.lib.recfunctions import (
    structured_to_unstructured,
    unstructured_to_structured,
)
from scipy import ndimage, optimize

try:
    from pyfftw.interfaces.numpy_fft import fftn as np_fftn
except ImportError:
    from numpy.fft import fftn as np_fftn

from pde.fields import ScalarField
from pde.grids import CylindricalSymGrid
from pde.grids.base import GridBase
from pde.grids.cartesian import CartesianGrid
from pde.grids.spherical import SphericalSymGridBase
from pde.tools.math import SmoothData1D
from pde.tools.typing import NumberOrArray

from .droplets import (
    DiffuseDroplet,
    PerturbedDroplet2D,
    PerturbedDroplet3D,
    SphericalDroplet,
)
from .emulsions import Emulsion


def threshold_otsu(data: np.ndarray, nbins: int = 256) -> float:
    """Find the threshold value for a bimodal histogram using the Otsu method.

    If you have a distribution that is bimodal, i.e., with two peaks and a valley
    between them, then you can use this to find the location of that valley, which
    splits the distribution into two.

    Args:
        data (:class:`~numpy.ndarray`):
            The data to be analyzed
        nbins (int):
            The number of bins in the histogram, which defines the accuracy of the
            determined threshold.

    Modified from https://stackoverflow.com/a/71345917/932593, which is based on the
    the SciKit Image threshold_otsu implementation:
    https://github.com/scikit-image/scikit-image/blob/70fa904eee9ef370c824427798302551df57afa1/skimage/filters/thresholding.py#L312
    """
    counts, bin_edges = np.histogram(data.flat, bins=nbins)
    bin_centers = (bin_edges[1:] + bin_edges[:-1]) / 2

    # class probabilities for all possible thresholds
    weight1 = np.cumsum(counts)
    weight2 = np.cumsum(counts[::-1])[::-1]
    # class means for all possible thresholds
    mean1 = np.cumsum(counts * bin_centers) / weight1
    mean2 = (np.cumsum((counts * bin_centers)[::-1]) / weight2[::-1])[::-1]

    # Clip ends to align class 1 and class 2 variables:
    # The last value of ``weight1``/``mean1`` should pair with zero values in
    # ``weight2``/``mean2``, which do not exist.
    variance12 = weight1[:-1] * weight2[1:] * (mean1[:-1] - mean2[1:]) ** 2

    idx = np.argmax(variance12)
    return float(bin_centers[idx])


def _locate_droplets_in_mask_cartesian(
    grid: CartesianGrid, mask: np.ndarray
) -> Emulsion:
    """locate droplets in a (potentially periodic) data set on a Cartesian grid

    This function locates droplets respecting periodic boundary conditions.

    Args:
        mask (:class:`~numpy.ndarray`):
            The binary image (or mask) in which the droplets are searched

    Returns:
        :class:`droplets.emulsions.Emulsion`: The discovered spherical droplets
    """
    if mask.shape != grid.shape:
        raise ValueError(
            f"The shape {mask.shape} of the data is not compatible with the grid "
            f"shape {grid.shape}"
        )

    # pad the array to simulate periodic boundary conditions
    offset = np.array([dim if p else 0 for p, dim in zip(grid.periodic, grid.shape)])
    pad = np.c_[offset, offset].astype(np.intc)
    mask_padded = np.pad(mask, pad, mode="wrap")
    assert np.all(mask_padded.shape == np.array(grid.shape) + 2 * offset)

    # locate individual clusters in the padded image
    labels, num_labels = ndimage.label(mask_padded)
    if num_labels == 0:
        return Emulsion([], grid=grid)
    indices = range(1, num_labels + 1)

    # create and emulsion from this of droplets
    grid._logger.info(f"Found {num_labels} droplet candidate(s)")

    # determine position from binary image and scale it to real space
    positions = ndimage.center_of_mass(mask_padded, labels, index=indices)
    # correct for the additional padding of the array
    positions = grid.transform(positions - offset, "cell", "grid")

    # determine volume from binary image and scale it to real space
    volumes = ndimage.sum(mask_padded, labels, index=indices)
    volumes = np.asanyarray(volumes) * np.prod(grid.discretization)

    # only retain droplets that are inside the central area
    droplets = (
        SphericalDroplet.from_volume(position, volume)
        for position, volume in zip(positions, volumes)
        if grid.cuboid.contains_point(position)
    )

    # filter overlapping droplets (e.g. due to duplicates)
    emulsion = Emulsion(droplets, grid=grid)
    num_candidates = len(emulsion)
    if num_candidates < num_labels:
        grid._logger.info(f"Only {num_candidates} candidate(s) inside bounds")

    emulsion.remove_overlapping()
    if len(emulsion) < num_candidates:
        grid._logger.info(f"Only {num_candidates} candidate(s) not overlapping")

    return emulsion


def _locate_droplets_in_mask_spherical(
    grid: SphericalSymGridBase, mask: np.ndarray
) -> Emulsion:
    """locates droplets in a binary data set on a spherical grid

    Args:
        mask (:class:`~numpy.ndarray`):
            The binary image (or mask) in which the droplets are searched

    Returns:
        :class:`droplets.emulsions.Emulsion`: The discovered spherical droplets
    """
    assert np.all(mask.shape == grid.shape)

    # locate clusters in the binary image
    labels, num_labels = ndimage.label(mask)
    if num_labels == 0:
        return Emulsion([], grid=grid)

    # locate clusters around origin
    object_slices = ndimage.find_objects(labels)
    droplet = None
    for slices in object_slices:
        if slices[0].start == 0:  # contains point around origin
            radius = float(grid.transform(slices[0].stop, "cell", "grid").flat[-1])
            droplet = SphericalDroplet(np.zeros(grid.dim), radius=radius)
        else:
            logger = logging.getLogger(grid.__class__.__module__)
            logger.warning("Found object not located at origin")

    # return an emulsion of droplets
    if droplet:
        return Emulsion([droplet], grid=grid)
    else:
        return Emulsion([], grid=grid)


def _locate_droplets_in_mask_cylindrical_single(
    grid: CylindricalSymGrid, mask: np.ndarray
) -> Emulsion:
    """locate droplets in a data set on a single cylindrical grid

    Args:
        mask (:class:`~numpy.ndarray`):
            The binary image (or mask) in which the droplets are searched

    Returns:
        :class:`droplets.emulsions.Emulsion`: The discovered spherical droplets
    """
    # locate the individual clusters
    labels, num_features = ndimage.label(mask)
    if num_features == 0:
        return Emulsion([], grid=grid)

    # locate clusters on the symmetry axis
    object_slices = ndimage.find_objects(labels)
    indices = []
    for index, slices in enumerate(object_slices, 1):
        if slices[0].start == 0:  # contains point on symmetry axis
            indices.append(index)
        else:
            logger = logging.getLogger(grid.__class__.__module__)
            logger.warning("Found object not located on symmetry axis")

    # determine position from binary image and scale it to real space
    pos = ndimage.center_of_mass(mask, labels, index=indices)
    pos = grid.transform(pos, "cell", "cartesian")

    # determine volume from binary image and scale it to real space
    vol_r, dz = grid.cell_volume_data
    cell_volumes = np.outer(vol_r, dz)
    try:
        vol = ndimage.sum_labels(cell_volumes, labels, index=indices)
    except AttributeError:
        # fall-back for older versions of scipy
        vol = ndimage.sum(cell_volumes, labels, index=indices)

    # return an emulsion of droplets
    droplets = (
        SphericalDroplet.from_volume(np.array([0, 0, p[2]]), v)
        for p, v in zip(pos, vol)
    )
    return Emulsion(droplets, grid=grid)


def _locate_droplets_in_mask_cylindrical(
    grid: CylindricalSymGrid, mask: np.ndarray
) -> Emulsion:
    """locate droplets in a data set on a (periodic) cylindrical grid

    This function locates droplets respecting periodic boundary conditions.

    Args:
        mask (:class:`~numpy.ndarray`):
            The binary image (or mask) in which the droplets are searched

    Returns:
        :class:`droplets.emulsions.Emulsion`: The discovered spherical droplets
    """
    assert np.all(mask.shape == grid.shape)

    if grid.periodic[1]:
        # locate droplets respecting periodic boundary conditions in z-direction

        # pad the array to simulate periodic boundary conditions
        dim_r, dim_z = grid.shape
        z_min, z_max = grid.axes_bounds[1]
        mask_padded = np.pad(mask, [[0, 0], [dim_z, dim_z]], mode="wrap")
        assert mask_padded.shape == (dim_r, 3 * dim_z)

        # locate droplets in the extended image
        candidates = _locate_droplets_in_mask_cylindrical_single(grid, mask_padded)
        grid._logger.info(f"Found {len(candidates)} droplet candidates.")

        # keep droplets that are inside the central area
        droplets = Emulsion(grid=grid)
        for droplet in candidates:
            # correct for the additional padding of the array
            droplet.position[2] -= grid.length
            # check whether the droplet lies in the original box
            if z_min <= droplet.position[2] <= z_max:
                droplets.append(droplet)

        grid._logger.info(f"Kept {len(droplets)} central droplets.")

        # filter overlapping droplets (e.g. due to duplicates)
        droplets.remove_overlapping()

    else:
        # simply locate droplets in the mask
        droplets = _locate_droplets_in_mask_cylindrical_single(grid, mask)

    return droplets


def locate_droplets_in_mask(grid: GridBase, mask: np.ndarray) -> Emulsion:
    """locates droplets in a binary image

    This function locates droplets respecting periodic boundary conditions.

    Args:
        mask (:class:`~numpy.ndarray`):
            The binary image (or mask) in which the droplets are searched

    Returns:
        :class:`droplets.emulsions.Emulsion`: The discovered spherical droplets
    """
    if isinstance(grid, CartesianGrid):
        return _locate_droplets_in_mask_cartesian(grid, mask)
    elif isinstance(grid, SphericalSymGridBase):
        return _locate_droplets_in_mask_spherical(grid, mask)
    elif isinstance(grid, CylindricalSymGrid):
        return _locate_droplets_in_mask_cylindrical(grid, mask)
    elif isinstance(grid, GridBase):
        raise NotImplementedError(f"Locating droplets is not possible for grid {grid}")
    else:
        raise ValueError(f"Invalid grid {grid}")


def locate_droplets(
    phase_field: ScalarField,
    threshold: Union[float, str] = 0.5,
    *,
    minimal_radius: float = 0,
    modes: int = 0,
    interface_width: Optional[float] = None,
    refine: bool = False,
    refine_args: Optional[Dict[str, Any]] = None,
) -> Emulsion:
    """Locates droplets in the phase field

    This uses a binarized image to locate clusters of large concentration in the
    phase field, which are interpreted as droplets. Basic quantities, like
    position and size, are determined for these clusters.

    Example:
        To determine the position, radius and interfacial width of an arbitrary droplet,
        the following call can be used

        .. code-block:: python

            emulsion = droplets.locate_droplets(
                field,
                threshold="auto",
                refine=True,
                refine_args={'vmin': None, 'vmax': None},
            )

        :code:`field` is the scalar field, in which the droplets are located. The
        `refine_args` set flexibel boundaries for the intensities inside and outside
        the droplet.

    Args:
        phase_field (:class:`~pde.fields.ScalarField`):
            Scalar field that describes the concentration field of droplets
        threshold (float or str):
            The threshold for binarizing the image. If a value is given it is used
            directly. Otherwise, the following algorithms are supported:

            * `extrema`: take mean between the minimum and the maximum of the data
            * `mean`: take the mean over the entire data
            * `otsu`: use Otsu's method implemented in :func:`threshold_otsu`

            The special value `auto` currently defaults to the `extrema` method.

        minimal_radius (float):
            The smallest radius of droplets to include in the list. This can be used to
            filter out fluctuations in noisy simulations.
        modes (int):
            The number of perturbation modes that should be included. If `modes=0`,
            droplets are assumed to be spherical. Note that the mode amplitudes are only
            determined when `refine=True`.
        interface_width (float, optional):
            Interface width of the located droplets, which is also used as a starting
            value for the fitting if droplets are refined.
        refine (bool):
            Flag determining whether the droplet properties should be refined using
            fitting. This is a potentially slow procedure.
        refine_args (dict):
            Additional keyword arguments passed on to :func:`refine_droplet`. Only has
            an effect if `refine=True`.

    Returns:
        :class:`~droplets.emulsions.Emulsion`: All detected droplets
    """
    assert isinstance(phase_field, ScalarField)
    dim = phase_field.grid.dim  # dimensionality of the space

    if modes > 0 and dim not in [2, 3]:
        raise ValueError("Perturbed droplets only supported for 2d and 3d")
    if refine_args is None:
        refine_args = {}

    # determine actual threshold
    if threshold == "extrema" or threshold == "auto":
        threshold = float(phase_field.data.min() + phase_field.data.max()) / 2
    elif threshold == "mean":
        threshold = float(phase_field.data.mean())
    elif threshold == "otsu":
        threshold = threshold_otsu(phase_field.data)
    else:
        threshold = float(threshold)

    # locate droplets in thresholded image
    img_binary = phase_field.data > threshold
    candidates = locate_droplets_in_mask(phase_field.grid, img_binary)

    if minimal_radius > -np.inf:
        candidates.remove_small(minimal_radius)

    droplets = []
    for droplet in candidates:
        # check whether we need to add the interface width
        droplet_class = droplet.__class__
        args: Dict[str, NumberOrArray] = {}

        # change droplet class when interface width is given
        if interface_width is not None:
            droplet_class = DiffuseDroplet
            args["interface_width"] = interface_width

        # change droplet class when perturbed droplets are requested
        if modes > 0:
            if dim == 2:
                droplet_class = PerturbedDroplet2D
            elif dim == 3:
                droplet_class = PerturbedDroplet3D
            else:
                raise NotImplementedError(f"Dimension {dim} is not supported")
            args["amplitudes"] = np.zeros(modes)

        # recreate a droplet of the correct class
        if droplet_class != droplet.__class__:
            droplet = droplet_class.from_droplet(droplet, **args)

        # refine droplets if necessary
        if refine:
            try:
                droplet = refine_droplet(phase_field, droplet, **refine_args)
            except ValueError:
                continue  # do not add the droplet to the list
        droplets.append(droplet)

    # return droplets as an emulsion
    emulsion = Emulsion(droplets, grid=phase_field.grid)
    if minimal_radius > -np.inf:
        emulsion.remove_small(minimal_radius)
    return emulsion


def refine_droplet(
    phase_field: ScalarField,
    droplet: DiffuseDroplet,
    *,
    vmin: float = 0.0,
    vmax: float = 1.0,
    adjust_values: bool = False,
    least_squares_params: Optional[Dict[str, Any]] = None,
) -> DiffuseDroplet:
    """Refines droplet parameters by fitting to phase field

    This function varies droplet parameters, like position, size,
    interface width, and potential perturbation amplitudes until the overlap
    with the respective phase field region is maximized. Here, we use a
    constraint fitting routine.

    Args:
        phase_field (:class:`~pde.fields.ScalarField`):
            Phase_field that is being used to refine the droplet
        droplet (:class:`~droplets.droplets.SphericalDroplet`):
            Droplet that should be refined. This could also be a subclass of
            :class:`SphericalDroplet`
        vmin (float):
            The intensity value of the dilute phase surrounding the droplet. If `None`,
            the value will be determined automatically.
        vmax (float):
            The intensity value inside the droplet. If `None`, the value will be
            determined automatically.
        adjust_value (bool):
            Flag determining whether the intensity values will be included in the
            fitting procedure. The default value `False` implies that the intensity
            values are regarded fixed.
        least_squares_params (dict):
            Dictionary of parameters that influence the fitting; see the documentation
            of :func:`scipy.optimize.least_squares`.

    Returns:
        :class:`droplets.droplets.DiffuseDroplet`: The refined droplet as an instance of
        the argument `droplet`
    """
    assert isinstance(phase_field, ScalarField)
    if least_squares_params is None:
        least_squares_params = {}

    if not isinstance(droplet, DiffuseDroplet):
        droplet = DiffuseDroplet.from_droplet(droplet)
    if droplet.interface_width is None:
        droplet.interface_width = phase_field.grid.typical_discretization

    # enlarge the mask to also contain the shape change
    mask = droplet._get_phase_field(phase_field.grid, dtype=np.bool_)
    dilation_iterations = 1 + int(2 * droplet.interface_width)
    mask = ndimage.binary_dilation(mask, iterations=dilation_iterations)

    # apply the mask
    data_mask = phase_field.data[mask]

    # determine the coordinate constraints and only vary the free data points
    data_flat = structured_to_unstructured(droplet.data)  # unstructured data
    dtype = droplet.data.dtype
    free = np.ones(len(data_flat), dtype=np.bool_)
    free[phase_field.grid.coordinate_constraints] = False

    # determine data bounds
    l, h = droplet.data_bounds
    bounds = l[free], h[free]

    # determine the intensities outside and inside the droplet
    if vmin is None:
        vmin = np.min(data_mask)
    if vmax is None:
        vmax = np.max(data_mask)
    vrng = vmax - vmin

    if adjust_values:
        # fit intensities in addition to all droplet parameters

        # add vmin and vrng as separate fitting parameters
        parameters = np.r_[data_flat[free], vmin, vmax]
        bounds = np.r_[bounds[0], vmin - vrng, 0], np.r_[bounds[1], vmax, 3 * vrng]

        def _image_deviation(params):
            """helper function evaluating the residuals"""
            # generate the droplet
            data_flat[free] = params[:-2]
            vmin, vrng = params[-2:]
            droplet.data = unstructured_to_structured(data_flat, dtype=dtype)
            droplet.check_data()
            img = vmin + vrng * droplet._get_phase_field(phase_field.grid)[mask]
            return img - data_mask

        # do the least square optimization
        result = optimize.least_squares(
            _image_deviation, parameters, bounds=bounds, **least_squares_params
        )
        data_flat[free] = result.x[:-2]

    else:
        # fit only droplet parameters and assume all intensities fixed

        def _image_deviation(params):
            """helper function evaluating the residuals"""
            # generate the droplet
            data_flat[free] = params
            droplet.data = unstructured_to_structured(data_flat, dtype=dtype)
            droplet.check_data()
            img = vmin + vrng * droplet._get_phase_field(phase_field.grid)[mask]
            return img - data_mask

        # do the least square optimization
        result = optimize.least_squares(
            _image_deviation, data_flat[free], bounds=bounds, **least_squares_params
        )
        data_flat[free] = result.x
    droplet.data = unstructured_to_structured(data_flat, dtype=dtype)

    # normalize the droplet position
    grid = phase_field.grid
    coords = grid.transform(droplet.position, "cartesian", "grid")
    droplet.position = grid.transform(grid.normalize_point(coords), "grid", "cartesian")

    return droplet


def get_structure_factor(
    scalar_field: ScalarField,
    smoothing: Union[None, float, str] = "auto",
    wave_numbers: Union[Sequence[float], str] = "auto",
    add_zero: bool = False,
) -> Tuple[np.ndarray, np.ndarray]:
    r"""Calculates the structure factor associated with a field

    Here, the structure factor is basically the power spectral density of the
    field `scalar_field` normalized so that re-gridding or rescaling the field
    does not change the result.

    Args:
        scalar_field (:class:`~pde.fields.ScalarField`):
            The scalar_field being analyzed
        smoothing (float, optional):
            Length scale that determines the smoothing of the radially averaged
            structure factor. If omitted, the full data about the discretized
            structure factor is returned. The special value `auto` calculates
            a value automatically.
        wave_numbers (list of floats, optional):
            The magnitude of the wave vectors at which the structure factor is
            evaluated. This only applies when smoothing is used. If `auto`, the
            wave numbers are determined automatically.
        add_zero (bool):
            Determines whether the value at k=0 (defined to be 1) should also be
            returned.

    Returns:
        (numpy.ndarray, numpy.ndarray): Two arrays giving the wave numbers and the
        associated structure factor. Wave numbers :math:`k` are related to distances by
        :math:`2\pi/k`.
    """
    logger = logging.getLogger(__name__)

    if not isinstance(scalar_field, ScalarField):
        raise TypeError(
            "Length scales can only be calculated for scalar "
            f"fields, not {scalar_field.__class__.__name__}"
        )

    grid = scalar_field.grid
    if not isinstance(grid, CartesianGrid):
        raise NotImplementedError(
            "Structure factor can currently only be calculated for Cartesian grids"
        )
    if not all(grid.periodic):
        logger.warning(
            "Structure factor calculation assumes periodic boundary "
            "conditions, but not all grid dimensions are periodic"
        )

    # do the n-dimensional Fourier transform and calculate the structure factor
    f1 = np_fftn(scalar_field.data, norm="ortho").flat[1:]
    flat_data = scalar_field.data.flat
    sf = np.abs(f1) ** 2 / np.dot(flat_data, flat_data)

    # an alternative calculation of the structure factor is
    #    f2 = np_ifftn(scalar_field.data, norm='ortho').flat[1:]
    #    sf = (f1 * f2).real
    #    sf /= (scalar_field.data**2).sum()
    # but since this involves two FFT, it is probably slower

    # determine the (squared) components of the wave vectors.
    # Note that `fftfreq` defines the wave number in cycles per unit of the sample
    # spacing, so we need to scale lengths by one over 2π.
    k2s = [
        np.fft.fftfreq(grid.shape[i], d=grid.discretization[i] / (2 * np.pi)) ** 2
        for i in range(grid.dim)
    ]
    # calculate the magnitude
    k_mag = np.sqrt(reduce(np.add.outer, k2s)).flat[1:]

    no_wavenumbers = wave_numbers is None or (
        isinstance(wave_numbers, str) and wave_numbers == "auto"
    )

    if smoothing is not None and smoothing != "none":
        # construct the smoothed function of the structure factor
        if smoothing == "auto":
            smoothing = k_mag.max() / 128
        smoothing = float(smoothing)  # type: ignore
        sf_smooth = SmoothData1D(k_mag, sf, sigma=smoothing)

        if no_wavenumbers:
            # determine the wave numbers at which to evaluate it
            k_min = 2 / grid.cuboid.size.max()
            k_max = k_mag.max()
            k_mag = np.linspace(k_min, k_max, 128)

        else:
            k_mag = np.array(wave_numbers)

        # obtain the smoothed values at these points
        sf = sf_smooth(k_mag)

    elif not no_wavenumbers:
        logger.warning(
            "Argument `wave_numbers` is only used when `smoothing` is enabled."
        )

    if add_zero:
        sf = np.r_[1, sf]
        k_mag = np.r_[0, k_mag]

    return k_mag, sf


def get_length_scale(
    scalar_field: ScalarField, method: str = "structure_factor_maximum", **kwargs
) -> Union[float, Tuple[float, Any]]:
    """Calculates a length scale associated with a phase field

    Args:
        scalar_field (:class:`~pde.fields.ScalarField`):
            The scala field being analyzed
        method (str):
            A string determining which method is used to calculate the length scale.
            Valid options are `structure_factor_maximum` (numerically determine the
            maximum in the structure factor) and `structure_factor_mean` (calculate the
            mean of the structure factor).

    Additional supported keyword arguments depend on the chosen method. For instance,
    the methods involving the structure factor allow for a boolean flag `full_output`,
    which also returns the actual structure factor. The method
    `structure_factor_maximum` also allows for some smoothing of the radially averaged
    structure factor. If the parameter `smoothing` is set to `None` the amount of
    smoothing is determined automatically from the typical discretization of the
    underlying grid. For the method `droplet_detection`, additional arguments are
    forwarded to :func:`locate_droplets`.

    Returns:
        float: The determine length scale

        If `full_output = True`, a tuple with the length scale and an additional
        object with further information is returned.

    See Also:
        :class:`~droplets.trackers.LengthScaleTracker`: Tracker measuring length scales
    """
    logger = logging.getLogger(__name__)

    if method == "structure_factor_mean" or method == "structure_factor_average":
        # calculate the structure factor
        k_mag, sf = get_structure_factor(scalar_field)
        length_scale = 2 * np.pi * np.sum(sf) / np.sum(k_mag * sf)

        if kwargs.pop("full_output", False):
            return length_scale, sf

    elif method == "structure_factor_maximum" or method == "structure_factor_peak":
        # calculate the structure factor
        k_mag, sf = get_structure_factor(scalar_field, smoothing=None, add_zero=True)

        # smooth the structure factor
        if kwargs.pop("smoothing", None) is None:
            smoothing = 0.01 * scalar_field.grid.typical_discretization
        sf_smooth = SmoothData1D(k_mag, sf, sigma=smoothing)

        # find the maximum
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            # determine maximum (excluding k=0)
            max_est = k_mag[1 + np.argmax(sf[1:])]
            bracket = np.array([0.2, 1, 5]) * max_est
            logger.debug(f"Search maximum of structure factor in interval {bracket}")
            try:
                result = optimize.minimize_scalar(
                    lambda x: -sf_smooth(x), bracket=bracket
                )
            except Exception:
                logger.exception("Could not determine maximal structure factor")
                length_scale = math.nan
            else:
                if not result.success:
                    logger.warning(
                        "Maximization of structure factor resulted in the following "
                        f"message: {result.message}"
                    )
                length_scale = 2 * np.pi / result.x

        if kwargs.pop("full_output", False):
            return length_scale, sf_smooth

    elif method == "droplet_detection":
        # calculate the length scale from detected droplets
        droplets = locate_droplets(scalar_field, **kwargs)
        kwargs = {}  # clear kwargs, so no warning is raised

        # get the axes along which droplets can be placed
        grid = scalar_field.grid
        axes = set(range(grid.dim)) - set(grid.coordinate_constraints)
        volume = 1.0
        for ax in axes:
            volume *= grid.axes_bounds[ax][1] - grid.axes_bounds[ax][0]

        volume_per_droplet = volume / len(droplets)
        length_scale = volume_per_droplet ** (1 / len(axes))

    else:
        raise ValueError(
            f"Method {method} is not defined. Valid values are `structure_factor_mean` "
            "and `structure_factor_maximum`"
        )

    if kwargs:
        # raise warning if keyword arguments remain
        logger.warning("Unused keyword arguments: %s", ", ".join(kwargs))

    # return only the length scale with out any additional information
    return length_scale  # type: ignore


__all__ = [
    "threshold_otsu",
    "locate_droplets",
    "refine_droplet",
    "get_structure_factor",
    "get_length_scale",
]



"""
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

import math
import random

import numpy as np
import pytest
from numba import njit
from scipy import integrate

from droplets.tools import spherical


@pytest.mark.parametrize("dim", [1, 2, 3])
def test_volume_conversion(dim):
    """tests conversion of volume and radius of droplet"""
    radius = 1 + random.random()
    volume = spherical.volume_from_radius(radius, dim=dim)
    radius2 = spherical.radius_from_volume(volume, dim=dim)
    assert radius2 == pytest.approx(radius)

    r2v = spherical.make_volume_from_radius_compiled(dim)
    v2r = spherical.make_radius_from_volume_compiled(dim)
    assert r2v(radius) == pytest.approx(volume)
    assert v2r(volume) == pytest.approx(radius)

    r2v_jit = njit(lambda r: r2v(r))
    v2r_jit = njit(lambda r: v2r(r))
    assert r2v_jit(radius) == pytest.approx(volume)
    assert v2r_jit(volume) == pytest.approx(radius)


@pytest.mark.parametrize("dim", [1, 2, 3])
def test_surface(dim):
    """test whether the surface is calculated correctly"""
    radius = 1 + random.random()
    eps = 1e-10
    vol1 = spherical.volume_from_radius(radius + eps, dim=dim)
    vol0 = spherical.volume_from_radius(radius, dim=dim)
    surface_approx = (vol1 - vol0) / eps
    surface = spherical.surface_from_radius(radius, dim=dim)
    assert surface == pytest.approx(surface_approx, rel=1e-3)

    r2s = spherical.make_surface_from_radius_compiled(dim)
    assert r2s(radius) == pytest.approx(surface)
    r2s_jit = njit(lambda r: r2s(r))
    assert r2s_jit(radius) == pytest.approx(surface)

    if dim == 1:
        with pytest.raises(RuntimeError):
            spherical.radius_from_surface(surface, dim=dim)
    else:
        assert spherical.radius_from_surface(surface, dim=dim) == pytest.approx(radius)


def test_spherical_conversion():
    """test the conversion between spherical and Cartesian coordinates"""
    s2c = spherical.points_spherical_to_cartesian
    c2s = spherical.points_cartesian_to_spherical

    ps = np.random.randn(64, 3)
    np.testing.assert_allclose(s2c(c2s(ps)), ps)

    # enforce angles
    ps[:, 0] = np.abs(ps[:, 0])  # radius is positive
    ps[:, 1] %= np.pi  # θ is between 0 and pi
    ps[:, 2] %= 2 * np.pi  # φ is between 0 and 2 pi
    np.testing.assert_allclose(c2s(s2c(ps)), ps, rtol=1e-6)


def test_spherical_index():
    """test the conversion of the spherical index"""
    # check initial state
    assert spherical.spherical_index_lm(0) == (0, 0)
    assert spherical.spherical_index_k(0, 0) == 0

    # check conversion
    for k in range(20):
        l, m = spherical.spherical_index_lm(k)
        assert spherical.spherical_index_k(l, m) == k

    # check order
    k = 0
    for l in range(4):
        for m in range(-l, l + 1):
            assert spherical.spherical_index_k(l, m) == k
            k += 1

    for l in range(4):
        k_max = spherical.spherical_index_k(l, l)
        assert spherical.spherical_index_count(l) == k_max + 1

    for l in range(4):
        for m in range(-l, l + 1):
            is_optimal = m == l
            k = spherical.spherical_index_k(l, m)
            assert spherical.spherical_index_count_optimal(k + 1) == is_optimal


def test_spherical_harmonics_real():
    """test spherical harmonics"""
    # test real spherical harmonics for symmetric case
    for deg in range(4):
        for _ in range(5):
            θ = math.pi * random.random()
            φ = 2 * math.pi * random.random()
            y1 = spherical.spherical_harmonic_symmetric(deg, θ)
            y2 = spherical.spherical_harmonic_real(deg, 0, θ, φ)
            assert y1 == y2

    # test orthogonality of real spherical harmonics
    deg = 1
    Ylm = spherical.spherical_harmonic_real
    for m1 in range(-deg, deg + 1):
        for m2 in range(-deg, m1 + 1):

            def integrand(t, p):
                return Ylm(deg, m1, t, p) * Ylm(deg, m2, t, p) * np.sin(t)

            overlap = integrate.dblquad(
                integrand, 0, 2 * np.pi, lambda _: 0, lambda _: np.pi
            )[0]
            if m1 == m2:
                assert overlap == pytest.approx(1)
            else:
                assert overlap == pytest.approx(0)



"""
This file is used to configure the test environment when running py.test

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

import matplotlib.pyplot as plt
import numpy as np
import pytest


@pytest.fixture(scope="function", autouse=True)
def suppressing():
    """helper function adjusting message reporting for all tests"""
    # raise all underflow errors
    np.seterr(all="raise", under="ignore")

    # run the actual test
    yield

    # clean up open matplotlib figures after the test
    plt.close("all")



"""
Simple sphinx plug-in that simplifies  type information in function signatures
"""

import collections
import re

# simple (literal) replacement rules
REPLACEMENTS = collections.OrderedDict(
    [
        # Remove some unnecessary information in favor of a more compact style
        ("Dict[KT, VT]", "dict"),
        ("Dict[str, Any]", "dict"),
        ("Optional[str]", "str"),
        ("Optional[float]", "float"),
        ("Optional[int]", "int"),
        ("Optional[dict]", "dict"),
        ("Optional[Dict[str, Any]]", "dict"),
        # Complex types that can be represented by some descriptive name
        ("Union[dict, str, BCBase]", "BoundaryData"),
        (
            "Union[Dict[str, BoundaryData], "
            "BoundaryData, Tuple[BoundaryData, BoundaryData]]",
            "BoundaryPairData",
        ),
        (
            "Union[Dict[str, BoundaryData], dict, str, BCBase, "
            "Tuple[BoundaryData, BoundaryData]]",
            "BoundaryPairData",
        ),
        ("Union[BoundaryPairData, Sequence[BoundaryPairData]]", "BoundariesData"),
        (
            "Union[Dict[str, BoundaryData], dict, str, BCBase, "
            "Tuple[BoundaryData, BoundaryData], Sequence[BoundaryPairData]]",
            "BoundariesData",
        ),
        (
            "Union[dict, str, BCBase, Tuple[Union[dict, str, BCBase], "
            "Union[dict, str, BCBase]], Sequence[Union[dict, str, BCBase, "
            "Tuple[Union[dict, str, BCBase], Union[dict, str, BCBase]]]]]",
            "BoundaryConditionData",
        ),
        (
            "Union[Dict[str, Union[Dict, str, BCBase]], Dict, str, BCBase, "
            "Tuple[Union[Dict, str, BCBase], Union[Dict, str, BCBase]], "
            "Sequence[Union[Dict[str, Union[Dict, str, BCBase]], Dict, str, BCBase, "
            "Tuple[Union[Dict, str, BCBase], Union[Dict, str, BCBase]]]]]",
            "BoundaryConditionData",
        ),
        (
            "Union[dict, str, BCBase, Tuple[Union[dict, str, BCBase], "
            "Union[dict, str, BCBase]], Sequence[Union[dict, str, BCBase, "
            "Tuple[Union[dict, str, BCBase], Union[dict, str, BCBase]]]], "
            "Sequence[BoundaryConditionData]]",
            "BoundaryConditionData",
        ),
        (
            "Union[Dict[str, BoundaryData], dict, str, BCBase, "
            "Tuple[BoundaryData, BoundaryData], "
            "Sequence[BoundaryPairData], Sequence[BoundariesData]]",
            "BoundariesDataList",
        ),
        (
            "Union[Dict[str, Union[Dict, str, BCBase]], Dict, str, BCBase, "
            "Tuple[Union[Dict, str, BCBase], Union[Dict, str, BCBase]], "
            "Sequence[Union[Dict[str, Union[Dict, str, BCBase]], Dict, str, BCBase, "
            "Tuple[Union[Dict, str, BCBase], Union[Dict, str, BCBase]]]], "
            "Sequence[BoundaryConditionData]]",
            "BoundariesDataList",
        ),
        ("Union[List[Union[TrackerBase, str]], TrackerBase, str, None]", "TrackerData"),
        (
            "Union[Generator, Sequence[droplets.droplets.SphericalDroplet]]",
            "DropletSequence",
        ),
    ]
)


# replacement rules based on regular expressions
REPLACEMENTS_REGEX = {
    # remove full package path and only leave the module/class identifier
    "pde\.(\w+\.)*": "",
}


def process_signature(
    app, what: str, name: str, obj, options, signature, return_annotation
):
    """Process signature by applying replacement rules"""
    if signature is not None:
        for key, value in REPLACEMENTS.items():
            signature = signature.replace(key, value)
        for key, value in REPLACEMENTS_REGEX.items():
            signature = re.sub(key, value, signature)
    return signature, return_annotation


def setup(app):
    """set up hooks for this sphinx plugin"""
    app.connect("autodoc-process-signature", process_signature)



"""
Functions and classes for analyzing emulsions and droplets

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

# determine the package version
try:
    # try reading version of the automatically generated module
    from ._version import __version__
except ImportError:
    # determine version automatically from CVS information
    from importlib.metadata import PackageNotFoundError, version

    try:
        __version__ = version("droplets")
    except PackageNotFoundError:
        # package is not installed, so we cannot determine any version
        __version__ = "unknown"
    del PackageNotFoundError, version  # clean name space

from .droplet_tracks import DropletTrack, DropletTrackList
from .droplets import DiffuseDroplet, SphericalDroplet
from .emulsions import Emulsion, EmulsionTimeCourse
from .image_analysis import get_length_scale, get_structure_factor, locate_droplets
from .trackers import DropletTracker, LengthScaleTracker



#!/usr/bin/env python3

import numpy as np

from droplets import DiffuseDroplet, Emulsion

# create 10 random droplets
droplets = [
    DiffuseDroplet(
        position=np.random.uniform(0, 100, 2),
        radius=np.random.uniform(5, 10),
        interface_width=1,
    )
    for _ in range(10)
]

# remove overlapping droplets in emulsion and plot it
emulsion = Emulsion(droplets)
emulsion.remove_overlapping()
emulsion.plot(color_value=lambda droplet: droplet.radius, colorbar="Droplet radius")



r"""
Module collecting functions for handling spherical geometry

The coordinate systems use the following convention for polar coordinates
:math:`(r, \phi)`, where :math:`r` is the radial coordinate and :math:`\phi` is
the polar angle:

.. math::
    \begin{cases}
        x = r \cos(\phi) &\\
        y = r \sin(\phi) &
    \end{cases}
    \text{for} \; r \in [0, \infty] \;
    \text{and} \; \phi \in [0, 2\pi)

Similarly, for spherical coordinates :math:`(r, \theta, \phi)`, where :math:`r`
is the radial coordinate, :math:`\theta` is the azimuthal angle, and
:math:`\phi` is the polar angle, we use

.. math::
    \begin{cases}
        x = r \sin(\theta) \cos(\phi) &\\
        y = r \sin(\theta) \sin(\phi) &\\
        z = r \cos(\theta)
    \end{cases}
    \text{for} \; r \in [0, \infty], \;
    \theta \in [0, \pi], \; \text{and} \;
    \phi \in [0, 2\pi)


The module also provides functions for handling spherical harmonics.
These spherical harmonics are described by the degree :math:`l` and the order 
:math:`m` or, alternatively, by the mode :math:`k`. The relation between these
values is

.. math::
    k = l(l + 1) + m

and

.. math::
    l &= \text{floor}(\sqrt{k}) \\
    m &= k - l(l + 1)

We will use these indices interchangeably, although the mode :math:`k` is
preferred internally. Note that we also consider axisymmetric spherical
harmonics, where the order is always zero and the degree :math:`l` and the mode
:math:`k` are thus identical.    


.. autosummary::
   :nosignatures:

   radius_from_volume
   volume_from_radius
   surface_from_radius
   radius_from_surface
   make_radius_from_volume_compiled
   make_volume_from_radius_compiled
   make_surface_from_radius_compiled
   points_cartesian_to_spherical
   points_spherical_to_cartesian
   spherical_index_k
   spherical_index_lm
   spherical_index_count
   spherical_index_count_optimal
   spherical_harmonic_symmetric
   spherical_harmonic_real
   spherical_harmonic_real_k


.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>    
"""

from typing import Callable, Tuple, TypeVar

import numpy as np
from numba.extending import overload, register_jitable
from scipy.special import sph_harm

from pde.grids.spherical import volume_from_radius
from pde.tools.numba import jit

π = float(np.pi)

TNumArr = TypeVar("TNumArr", float, np.ndarray)


def radius_from_volume(volume: TNumArr, dim: int) -> TNumArr:
    """Return the radius of a sphere with a given volume

    Args:
        volume (float or :class:`~numpy.ndarray`): Volume of the sphere
        dim (int): Dimension of the space

    Returns:
        float or :class:`~numpy.ndarray`: Radius of the sphere
    """
    if dim == 1:
        return volume / 2
    elif dim == 2:
        return np.sqrt(volume / π)  # type: ignore
    elif dim == 3:
        return (3 * volume / (4 * π)) ** (1 / 3)  # type: ignore
    else:
        raise NotImplementedError(f"Cannot calculate the radius in {dim} dimensions")


def make_radius_from_volume_compiled(dim: int) -> Callable[[TNumArr], TNumArr]:
    """Return a function calculating the radius of a sphere with a given volume

    Args:
        dim (int): Dimension of the space. If omitted, a general function is returned

    Returns:
        function: A function that takes a volume and returns the radius
    """
    if dim == 1:

        def radius_from_volume(volume: TNumArr) -> TNumArr:
            return volume / 2

    elif dim == 2:

        def radius_from_volume(volume: TNumArr) -> TNumArr:
            return np.sqrt(volume / π)  # type: ignore

    elif dim == 3:

        def radius_from_volume(volume: TNumArr) -> TNumArr:
            return (3 * volume / (4 * π)) ** (1 / 3)  # type: ignore

    else:
        raise NotImplementedError(f"Cannot calculate the radius in {dim} dimensions")
    return jit(radius_from_volume)  # type: ignore


def make_radius_from_volume_nd_compiled() -> Callable[[TNumArr, int], TNumArr]:
    """Return a function calculating the radius of a sphere with a given volume

    Returns:
        function: A function that calculate the radius from a volume and dimension
    """

    @register_jitable
    def radius_from_volume(volume: TNumArr, dim: int) -> TNumArr:
        if dim == 1:
            return volume / 2
        elif dim == 2:
            return np.sqrt(volume / π)  # type: ignore
        elif dim == 3:
            return (3 * volume / (4 * π)) ** (1 / 3)  # type: ignore
        raise NotImplementedError

    return radius_from_volume  # type: ignore


def make_volume_from_radius_compiled(dim: int) -> Callable[[TNumArr], TNumArr]:
    """Return a function calculating the volume of a sphere with a given radius

    Args:
        dim (int): Dimension of the space

    Returns:
        function: A function that takes a radius and returns the volume
    """
    if dim == 1:

        def volume_from_radius(radius: TNumArr) -> TNumArr:
            return 2 * radius

    elif dim == 2:

        def volume_from_radius(radius: TNumArr) -> TNumArr:
            return π * radius**2

    elif dim == 3:

        def volume_from_radius(radius: TNumArr) -> TNumArr:
            return 4 * π / 3 * radius**3

    else:
        raise NotImplementedError(f"Cannot calculate the volume in {dim} dimensions")
    return jit(volume_from_radius)  # type: ignore


def make_volume_from_radius_nd_compiled() -> Callable[[TNumArr, int], TNumArr]:
    """Return a function calculating the volume of a sphere with a given radius

    Returns:
        function: A function that calculates the volume using a radius and dimension
    """

    @register_jitable
    def volume_from_radius(radius: TNumArr, dim: int) -> TNumArr:
        if dim == 1:
            return 2 * radius
        elif dim == 2:
            return π * radius**2
        elif dim == 3:
            return 4 * π / 3 * radius**3
        raise NotImplementedError

    return volume_from_radius  # type: ignore


def surface_from_radius(radius: TNumArr, dim: int) -> TNumArr:
    """Return the surface area of a sphere with a given radius

    Args:
        radius (float or :class:`~numpy.ndarray`): Radius of the sphere
        dim (int): Dimension of the space

    Returns:
        float or :class:`~numpy.ndarray`: Surface area of the sphere
    """
    if dim == 1:
        if isinstance(radius, np.ndarray):
            return np.broadcast_to(2, radius.shape)
        else:
            return 2
    elif dim == 2:
        return 2 * π * radius
    elif dim == 3:
        return 4 * π * radius**2
    else:
        raise NotImplementedError(
            f"Cannot calculate the surface area in {dim} dimensions"
        )


def radius_from_surface(surface: TNumArr, dim: int) -> TNumArr:
    """Return the radius of a sphere with a given surface area

    Args:
        surface (float or :class:`~numpy.ndarray`): Surface area of the sphere
        dim (int): Dimension of the space

    Returns:
        float or :class:`~numpy.ndarray`: Radius of the sphere
    """
    if dim == 1:
        raise RuntimeError("Cannot calculate radius of 1-d sphere from surface")
    elif dim == 2:
        return surface / (2 * π)
    elif dim == 3:
        return np.sqrt(surface / (4 * π))  # type: ignore
    else:
        raise NotImplementedError(f"Cannot calculate the radius in {dim} dimensions")


def make_surface_from_radius_compiled(dim: int) -> Callable[[TNumArr], TNumArr]:
    """Return a function calculating the surface area of a sphere

    Args:
        dim (int): Dimension of the space

    Returns:
        function: A function that takes a radius and returns the surface area
    """
    import numba as nb

    if dim == 1:

        def _surface_from_radius(radius):
            if isinstance(radius, np.ndarray):
                return np.full(radius.shape, 2)
            else:
                return 2

        @overload(_surface_from_radius)
        def ol_surface_from_radius(radius):
            if isinstance(radius, nb.types.Array):
                return lambda radius: np.full(radius.shape, 2)
            else:
                return lambda radius: 2

        @jit
        def surface_from_radius(radius: TNumArr) -> TNumArr:
            return _surface_from_radius(radius)  # type: ignore

    elif dim == 2:

        @jit
        def surface_from_radius(radius: TNumArr) -> TNumArr:
            return 2 * π * radius

    elif dim == 3:

        @jit
        def surface_from_radius(radius: TNumArr) -> TNumArr:
            return 4 * π * radius**2

    else:
        raise NotImplementedError(
            f"Cannot calculate the surface area in {dim} dimensions"
        )
    return surface_from_radius  # type: ignore


def points_cartesian_to_spherical(points: np.ndarray) -> np.ndarray:
    """Convert points from Cartesian to spherical coordinates

    Args:
        points (:class:`~numpy.ndarray`): Points in Cartesian coordinates

    Returns:
        :class:`~numpy.ndarray`: Points (r, θ, φ) in spherical coordinates
    """
    points = np.atleast_1d(points)
    assert points.shape[-1] == 3, "Points must have 3 coordinates"

    ps_spherical = np.empty(points.shape)
    # calculate radius in [0, infinity]
    ps_spherical[..., 0] = np.linalg.norm(points, axis=-1)
    # calculate θ in [0, pi]
    ps_spherical[..., 1] = np.arccos(points[..., 2] / ps_spherical[..., 0])
    # calculate φ in [0, 2 * pi]
    ps_spherical[..., 2] = np.arctan2(points[..., 1], points[..., 0]) % (2 * π)
    return ps_spherical


def points_spherical_to_cartesian(points: np.ndarray) -> np.ndarray:
    """Convert points from spherical to Cartesian coordinates

    Args:
        points (:class:`~numpy.ndarray`):
            Points in spherical coordinates (r, θ, φ)

    Returns:
        :class:`~numpy.ndarray`: Points in Cartesian coordinates
    """
    points = np.atleast_1d(points)
    assert points.shape[-1] == 3, "Points must have 3 coordinates"

    sin_θ = np.sin(points[..., 1])
    ps_cartesian = np.empty(points.shape)
    ps_cartesian[..., 0] = points[..., 0] * np.cos(points[..., 2]) * sin_θ
    ps_cartesian[..., 1] = points[..., 0] * np.sin(points[..., 2]) * sin_θ
    ps_cartesian[..., 2] = points[..., 0] * np.cos(points[..., 1])
    return ps_cartesian


def spherical_index_k(degree: int, order: int = 0) -> int:
    """returns the mode `k` from the degree `degree` and order `order`

    Args:
        degree (int): Degree of the spherical harmonics
        order (int): Order of the spherical harmonics

    Raises:
        ValueError: if `order < -degree` or `order > degree`

    Returns:
        int: a combined index k
    """
    if not -degree <= order <= degree:
        raise ValueError("order must lie between -degree and degree")
    return degree * (degree + 1) + order


def spherical_index_lm(k: int) -> Tuple[int, int]:
    """returns the degree `l` and the order `m` from the mode `k`

    Args:
        k (int): The combined index for the spherical harmonics

    Returns:
        tuple: The degree `l` and order `m` of the spherical harmonics
        assoicated with the combined index
    """
    degree = int(np.floor(np.sqrt(k)))
    return degree, k - degree * (degree + 1)


def spherical_index_count(l: int) -> int:
    """return the number of modes for all indices <= l

    The returned value is one less than the maximal mode `k` required.

    Args:
        l (int): Maximal degree of the spherical harmonics

    Returns:
        int: The number of modes
    """
    return 1 + 2 * l + l * l


def spherical_index_count_optimal(k_count: int) -> bool:
    """checks whether the modes captures all orders for maximal degree

    Args:
        k_count (int): The number of modes considered
    """
    is_square = bool(int(np.sqrt(k_count) + 0.5) ** 2 == k_count)
    return is_square


def spherical_harmonic_symmetric(degree: int, θ: float) -> float:
    r"""axisymmetric spherical harmonics with degree `degree`, so `m=0`.

    Args:
        degree (int): Degree of the spherical harmonics
        θ (float): Azimuthal angle at which the spherical harmonics is
            evaluated (in :math:`[0, \pi]`)

    Returns:
        float: The value of the spherical harmonics
    """
    # note that the definition of `sph_harm` has a different convention for the
    # usage of the variables φ and θ and we thus have to swap the args
    return np.real(sph_harm(0.0, degree, 0.0, θ))  # type: ignore


def spherical_harmonic_real(degree: int, order: int, θ: float, φ: float) -> float:
    r"""real spherical harmonics of degree l and order m

    Args:
        degree (int): Degree :math:`l` of the spherical harmonics
        order (int): Order :math:`m` of the spherical harmonics
        θ (float): Azimuthal angle (in :math:`[0, \pi]`) at which the
            spherical harmonics is evaluated.
        φ (float): Polar angle (in :math:`[0, 2\pi]`) at which the spherical
            harmonics is evaluated.

    Returns:
        float: The value of the spherical harmonics
    """
    # note that the definition of `sph_harm` has a different convention for the
    # usage of the variables φ and θ and we thus have to swap the args
    # Moreover, the scipy functions expect first the order and then the degree
    if order > 0:
        term1 = sph_harm(order, degree, φ, θ)
        term2 = (-1) ** order * sph_harm(-order, degree, φ, θ)
        return np.real((term1 + term2) / np.sqrt(2))  # type: ignore

    elif order == 0:
        return np.real(sph_harm(0, degree, φ, θ))  # type: ignore

    else:  # order < 0
        term1 = sph_harm(-order, degree, φ, θ)
        term2 = (-1) ** order * sph_harm(order, degree, φ, θ)
        return np.real((term1 - term2) / (complex(0, np.sqrt(2))))  # type: ignore


def spherical_harmonic_real_k(k: int, θ: float, φ: float) -> float:
    r"""real spherical harmonics described by mode k

    Args:
        k (int): Combined index determining the degree and order of the
            spherical harmonics
        θ (float): Azimuthal angle (in :math:`[0, \pi]`) at which the
            spherical harmonics is evaluated.
        φ (float): Polar angle (in :math:`[0, 2\pi]`) at which the spherical
            harmonics is evaluated.

    Returns:
        float: The value of the spherical harmonics
    """
    return spherical_harmonic_real(*spherical_index_lm(k), θ=θ, φ=φ)



"""
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

import numpy as np
import pytest
from scipy import integrate

from pde.grids import UnitGrid
from pde.tools.misc import skipUnlessModule
from pde.tools.numba import jit

from droplets import droplets


def test_simple_droplet():
    """test a given simple droplet"""
    d = droplets.SphericalDroplet((1, 2), 1)
    assert d.surface_area == pytest.approx(2 * np.pi)
    np.testing.assert_allclose(d.interface_position(0), [2, 2])
    np.testing.assert_allclose(d.interface_position([0]), [[2, 2]])

    d.volume = 3
    assert d.volume == pytest.approx(3)


@pytest.mark.parametrize("dim", [1, 2, 3])
def test_random_droplet(dim):
    """tests simple droplet"""
    pos = np.random.uniform(0, 10, dim)
    radius = np.random.uniform(2, 3)
    d1 = droplets.SphericalDroplet(pos, radius)
    d2 = droplets.SphericalDroplet(np.zeros(dim), radius)
    d2.position = pos

    assert d1.dim == dim
    assert d1.volume > 0
    assert d1.surface_area > 0
    assert d1 == d2

    d3 = d1.copy()
    assert d1 == d3
    assert d1 is not d3

    vol = np.random.uniform(10, 30)
    d2.volume = vol
    assert d2.volume == pytest.approx(vol)

    f = d1.get_phase_field(UnitGrid([10] * dim), vmin=0.2, vmax=0.8, label="test")
    assert f.label == "test"
    assert np.all(f.data >= 0.2)
    assert np.all(f.data <= 0.8)
    assert np.any(f.data == 0.2)
    assert np.any(f.data == 0.8)


def test_perturbed_droplet_2d():
    """test methods of perturbed droplets in 2d"""
    d = droplets.PerturbedDroplet2D([0, 1], 1, 0.1, [0.0, 0.1, 0.2])
    d.volume
    d.interface_distance(0.1)
    d.interface_position(0.1)
    d.interface_curvature(0.1)


def test_perturbed_droplet_3d():
    """test methods of perturbed droplets in 2d"""
    d = droplets.PerturbedDroplet3D([0, 1, 2], 1, 0.1, [0.0, 0.1, 0.2, 0.3])
    d.volume_approx
    d.interface_distance(0.1, 0.2)
    d.interface_position(0.1, 0.2)
    d.interface_curvature(0.1, 0.2)


def test_perturbed_volume():
    """test volume calculation of perturbed droplets"""
    pos = np.random.randn(2)
    radius = 1 + np.random.random()
    amplitudes = np.random.uniform(-0.2, 0.2, 6)
    d = droplets.PerturbedDroplet2D(pos, radius, 0, amplitudes)

    def integrand(φ):
        r = d.interface_distance(φ)
        return 0.5 * r**2

    vol = integrate.quad(integrand, 0, 2 * np.pi)[0]
    assert vol == pytest.approx(d.volume)

    vol = np.random.uniform(1, 2)
    d.volume = vol
    assert vol == pytest.approx(d.volume)

    pos = np.random.randn(3)
    radius = 1 + np.random.random()
    d = droplets.PerturbedDroplet3D(pos, radius, 0, np.zeros(7))
    assert d.volume == pytest.approx(4 * np.pi / 3 * radius**3)


def test_surface_area():
    """test surface area calculation of droplets"""
    # perturbed 2d droplet
    R0 = 3
    amplitudes = np.random.uniform(-1e-2, 1e-2, 6)

    # unperturbed droplets
    d1 = droplets.SphericalDroplet([0, 0], R0)
    d2 = droplets.PerturbedDroplet2D([0, 0], R0)
    assert d1.surface_area == pytest.approx(d2.surface_area)
    assert d2.surface_area == pytest.approx(d2.surface_area_approx)

    # perturbed droplet
    d1 = droplets.SphericalDroplet([0, 0], R0)
    d2 = droplets.PerturbedDroplet2D([0, 0], R0, amplitudes=amplitudes)
    assert d1.surface_area != pytest.approx(d2.surface_area)
    assert d2.surface_area == pytest.approx(d2.surface_area_approx, rel=1e-4)


def test_curvature():
    """test interface curvature calculation"""
    # spherical droplet
    for dim in range(1, 4):
        d = droplets.SphericalDroplet(np.zeros(dim), radius=np.random.uniform(1, 4))
        assert d.interface_curvature == pytest.approx(1 / d.radius)

    # perturbed 2d droplet
    R0 = 3
    epsilon = 0.1
    amplitudes = epsilon * np.array([0.1, 0.2, 0.3, 0.4])

    def curvature_analytical(φ):
        """analytical expression for curvature"""
        radius = (
            3.0
            * (
                5.0 * (40.0 + 27.0 * epsilon**2.0)
                + epsilon
                * (
                    40.0 * (4.0 * np.cos(2.0 * φ) + np.sin(φ))
                    + np.cos(φ) * (80.0 + 66.0 * epsilon + 240.0 * np.sin(φ))
                    - epsilon
                    * (
                        10.0 * np.cos(3.0 * φ)
                        + 21.0 * np.cos(4.0 * φ)
                        - 12.0 * np.sin(φ)
                        + 20.0 * np.sin(3.0 * φ)
                        + 72.0 * np.sin(4.0 * φ)
                    )
                )
            )
            ** (3.0 / 2.0)
            / (
                10.0
                * np.sqrt(2.0)
                * (
                    200.0
                    + 60.0
                    * epsilon
                    * (
                        2.0 * np.cos(φ)
                        + 8.0 * np.cos(2.0 * φ)
                        + np.sin(φ)
                        + 6.0 * np.sin(2.0 * φ)
                    )
                    + epsilon**2.0
                    * (
                        345.0
                        + 165.0 * np.cos(φ)
                        - 5.0 * np.cos(3.0 * φ)
                        - 21.0 * np.cos(4.0 * φ)
                        + 30.0 * np.sin(φ)
                        - 10.0 * np.sin(3.0 * φ)
                        - 72.0 * np.sin(4.0 * φ)
                    )
                )
            )
        )
        return 1 / radius

    d = droplets.PerturbedDroplet2D([0, 0], R0, amplitudes=amplitudes)
    φs = np.linspace(0, np.pi, 64)
    np.testing.assert_allclose(
        d.interface_curvature(φs), curvature_analytical(φs), rtol=1e-1
    )


def test_from_data():
    """test the from_data constructor"""
    for d1 in [
        droplets.SphericalDroplet((1,), 2),
        droplets.SphericalDroplet((1, 2), 3),
        droplets.PerturbedDroplet2D((1, 2), 3, 0.1, [0.1, 0.2]),
        droplets.PerturbedDroplet3D((1, 2, 3), 4, 0.1, [0.1, 0.2]),
    ]:
        d2 = d1.__class__.from_data(d1.data)
        assert d1 == d2
        assert d1 is not d2


@skipUnlessModule("h5py")
def test_triangulation_2d():
    """test the 2d triangulation of droplets"""
    d1 = droplets.SphericalDroplet([1, 3], 5)
    d2 = droplets.PerturbedDroplet2D([2, 4], 5, amplitudes=[0.1, 0.2, 0.1, 0.2])
    for drop in [d1, d2]:
        tri = drop.get_triangulation(0.1)
        l = sum(
            np.linalg.norm(tri["vertices"][i] - tri["vertices"][j])
            for i, j in tri["lines"]
        )
        assert l == pytest.approx(drop.surface_area, rel=1e-3), drop


@skipUnlessModule("h5py")
def test_triangulation_3d():
    """test the 3d triangulation of droplets"""
    d1 = droplets.SphericalDroplet([1, 2, 3], 5)
    d2 = droplets.PerturbedDroplet3D([2, 3, 4], 5, amplitudes=[0.1, 0.2, 0.1, 0.2])
    for drop in [d1, d2]:
        tri = drop.get_triangulation(1)
        vertices = tri["vertices"]
        vol = 0
        for a, b, c in tri["triangles"]:
            # calculate the total volume by adding the volumes of the tetrahedra
            mat = np.c_[vertices[a], vertices[b], vertices[c], drop.position]
            mat = np.vstack([mat, np.ones(4)])
            vol += abs(np.linalg.det(mat) / 6)
        assert vol == pytest.approx(drop.volume, rel=0.1), drop


@pytest.mark.parametrize("cls", [droplets.SphericalDroplet, droplets.DiffuseDroplet])
@pytest.mark.parametrize("dim", [1, 2, 3])
def test_droplet_merge(cls, dim):
    """test merging of droplets"""
    if cls == droplets.SphericalDroplet:
        d1 = droplets.SphericalDroplet([0] * dim, 1)
        d2 = droplets.SphericalDroplet([2] * dim, 1)
    elif cls == droplets.DiffuseDroplet:
        d1 = droplets.DiffuseDroplet([0] * dim, 1, 1)
        d2 = droplets.DiffuseDroplet([2] * dim, 1, 2)

    # test simple merge
    d3 = d1.merge(d2, inplace=False)
    assert d3 is not d1 and d3 is not d2
    np.testing.assert_allclose(d3.position, [1] * dim)
    assert d3.volume == pytest.approx(d1.volume + d2.volume)
    if cls == droplets.DiffuseDroplet:
        assert d3.interface_width == pytest.approx(1.5)

    # test numba merge
    merge = jit(d1._make_merge_data())
    d3_data = d1.data.copy()
    merge(d1.data, d2.data, d3_data)
    np.testing.assert_allclose(d3.position, [1] * dim)
    assert d3_data.radius == pytest.approx(d3.radius)
    if cls == droplets.DiffuseDroplet:
        assert d3_data.interface_width == pytest.approx(1.5)

    # test inplace merge
    d4 = d1.merge(d2, inplace=True)
    assert d4 is d1 and d3 is not d2
    np.testing.assert_allclose(d4.position, [1] * dim)
    assert d4.volume == pytest.approx(d3.volume)
    if cls == droplets.DiffuseDroplet:
        assert d4.interface_width == pytest.approx(1.5)


def test_droplet_interface_merge():
    """test merging of droplets with diffuse interfaces"""
    d1 = droplets.DiffuseDroplet([0, 0], 1, 1)
    d2 = droplets.DiffuseDroplet([2, 2], 1, 2)
    d1.merge(d2, inplace=True)
    assert d1.interface_width == pytest.approx(1.5)
